"""Dataflow helpers."""

from sqlalchemy.orm import Query, aliased

from mate_common.models.cpg_types import EdgeKind, Opcode
from mate_query.cpg.models.core import Edge, Node
from mate_query.db import Graph as CPG


def memory_location_of_pointer_arg(
    Node_: Node, Edge_: Edge, call_uuid: str, argument_no: int
) -> Node:
    """Returns the Memory Location node pointed to by the argument_number'th argument of a function.

    If a function takes pointer arguments and modifies the memory those arguments point to
    then data flows through those pointers. To allow you to trace that dataflow, you need
    to find the memory locations of those pointer arguments, and then perform a dataflow query
    from there. This function returns the memory location of the operand_no'th pointer argument.

    Args:
        Node_: the Node sql object
        Edge_: the Edge sql object
        call_uuid (string)
        argument_no (int)

    Returns:
        Node_: the memory location from which to trace dataflow
    """
    e1 = aliased(Edge_)  # ValueDefinitionToUse
    e2 = aliased(Edge_)  # LoadMemory
    return (
        Query(Node_)
        .join(e2, Node_.uuid == e2.target)  # MemoryLocation -> load
        .join(e1, e1.source == e2.source)  # load -> call
        .filter(
            (e1.kind == EdgeKind.VALUE_DEFINITION_TO_USE)
            & (e2.kind == EdgeKind.POINTS_TO)
            & e1.target_node.has(Node_.attributes["opcode"].as_string() == Opcode.CALL.value)
            & (e1.target == call_uuid)
            & (e1.attributes["operand_number"].as_integer() == argument_no)
        )
    )


def get_pointer_arguments(cpg: CPG, function_name: str) -> Query:
    """Returns a query for all Argument nodes which are pointers.

    Args:
        Argument: the sql Argument object
        function_name (string)
    Returns:
        Query which can be evaluated for Argument node(s)
    """
    return Query(cpg.Argument).filter(
        cpg.Argument.parent_function.has(cpg.Function.name == function_name),
        cpg.Argument.llvm_type.has(cpg.LLVMType.is_pointer_type),
    )



"""Core edge models and functions.

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKeyConstraint, Index, Table, text
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import RelationshipProperty, backref, relationship
from sqlalchemy.types import Enum as SAEnum
from sqlalchemy.types import String

from mate_common.assertions import mate_assert
from mate_common.models.cpg_types.mate import EdgeKind
from mate_query.cpg.models.core.base import Base

if TYPE_CHECKING:
    from typing import Any, ClassVar, Dict, Optional, Tuple, Union

    from mate_query.cpg.models.core.node import BaseNode
    from mate_query.db import Graph as CPG


def edge_kind_column() -> Column:
    return Column(
        "kind",
        SAEnum(EdgeKind, values_callable=lambda x: [e.value for e in x], validate_strings=True),
        index=True,
        nullable=False,
    )


def source_column(_node_uuid_column: Optional[Column] = None) -> Column:
    return Column(
        "source",
        String,
        doc="The uuid of the edge's source node.",
        nullable=False,
    )


def target_column(_node_uuid_column: Optional[Column] = None) -> Column:
    return Column(
        "target",
        String,
        doc="The uuid of the edge's target node.",
        nullable=False,
    )


def edge_foreign_key_constraints(node_table: Table) -> Tuple[ForeignKeyConstraint, ...]:
    return (
        ForeignKeyConstraint(["source"], [node_table.c.uuid]),
        ForeignKeyConstraint(["target"], [node_table.c.uuid]),
    )


class BaseEdge(Base):
    # NOTE(lb): This is shared between SQLite (declarative) and Postgres
    # (imperative) mappings.
    # TODO(#1044): Remove me!

    cpg: ClassVar[CPG]

    # NOTE(lb): Why Node when we already have cpg.Node in scope? We map db.Node
    # and db.Edge without a CPG handy, so the ClassVar[CPG] is a bit of a fib
    # in that specific case.
    Node: ClassVar[BaseNode]

    kind = edge_kind_column()

    @declared_attr
    def __table_args__(cls) -> Tuple[Union[ForeignKeyConstraint, Index], ...]:
        args = list(edge_foreign_key_constraints(cls.Node.__table__))
        args.extend(
            [
                Index(
                    f"edge_{cls.build_id}_kind_source",
                    edge_kind_column(),
                    source_column(cls.Node.uuid),
                ),
                Index(
                    f"edge_{cls.build_id}_kind_target",
                    edge_kind_column(),
                    target_column(cls.Node.uuid),
                ),
                Index(
                    f"edge_{cls.build_id}_source_kind",
                    source_column(cls.Node.uuid),
                    edge_kind_column(),
                ),
                Index(
                    f"edge_{cls.build_id}_target_kind",
                    target_column(cls.Node.uuid),
                    edge_kind_column(),
                ),
                Index(
                    f"edge_{cls.build_id}_kind_source_context",
                    edge_kind_column(),
                    source_column(cls.Node.uuid),
                    text("(attributes->>'context')"),
                ),
                Index(
                    f"edge_{cls.build_id}_kind_target_context",
                    edge_kind_column(),
                    target_column(cls.Node.uuid),
                    text("(attributes->>'context')"),
                ),
                Index(
                    f"edge_{cls.build_id}_kind_source_caller",
                    edge_kind_column(),
                    source_column(cls.Node.uuid),
                    text("(attributes->>'caller_context')"),
                ),
                Index(
                    f"edge_{cls.build_id}_kind_target_callee",
                    edge_kind_column(),
                    target_column(cls.Node.uuid),
                    text("(attributes->>'callee_context')"),
                ),
            ]
        )
        return tuple(args)

    @declared_attr
    def source(cls) -> Column:
        return source_column(cls.Node.uuid)

    @declared_attr
    def target(cls) -> Column:
        return target_column(cls.Node.uuid)

    @declared_attr
    def source_node(cls) -> RelationshipProperty:
        return relationship(
            cls.Node,
            backref=backref("outgoing", doc="All edges where this node is the source."),
            foreign_keys=cls.source,
        )

    @declared_attr
    def target_node(cls) -> RelationshipProperty:
        return relationship(
            cls.Node,
            backref=backref("incoming", doc="All edges where this node is the target."),
            foreign_keys=cls.target,
        )

    def __repr__(self) -> str:
        return f"<Edge({self.uuid}: {self.source} -[{self.kind}]> {self.target})>"

    if TYPE_CHECKING:
        # Make Mypy happier. Unused by SQLAlchemy.
        def __init__(
            self,
            uuid: str,  # pylint: disable=unused-argument
            kind: EdgeKind,  # pylint: disable=unused-argument
            source: str,  # pylint: disable=unused-argument
            target: str,  # pylint: disable=unused-argument
            attributes: Dict[str, Any] = dict(),  # pylint: disable=unused-argument
        ):
            mate_assert(False, "Unused. Just exists to make Mypy happy.")


class Edge(BaseEdge):
    # Make Mypy happier. Unused by SQLAlchemy.
    def __init__(
        self,
        uuid: str,  # pylint: disable=unused-argument
        kind: EdgeKind,  # pylint: disable=unused-argument
        source: str,  # pylint: disable=unused-argument
        target: str,  # pylint: disable=unused-argument
        attributes: Dict[str, Any] = dict(),  # pylint: disable=unused-argument
    ):
        mate_assert(False, "Unused. Just exists to make Mypy happy.")



"""This package contains specializations of the ``Node`` class.

The query interface automatically instantiates these classes when the
appropriate type of node is retrieved from the database. For instance,
if a query returns a set of instructions, the result will be an iterable
of objects of the ``Instruction`` class, rather than the bare ``Node``
class.

In addition to accessing the attributes of these classes after a query is
executed, the attributes and relationships can also be used as SQLAlchemy
expressions in a filter, and the classes themselves can be used as if they were
tables.
"""
from .analysis import *
from .ast import *
from .dwarf import *
from .translation_unit import *



"""Analysis nodes.

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import and_, or_
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, relationship

from mate_common.models.cpg_types import EdgeKind, NodeKind
from mate_query.cpg.models.core.relationships import Direction
from mate_query.cpg.models.node._typechecking import NodeMixin

if TYPE_CHECKING:

    from sqlalchemy.orm import RelationshipProperty


class MemoryLocation(NodeMixin):
    _kind = NodeKind.MEMORY_LOCATION

    @declared_attr
    def loaded_to(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.LOAD_MEMORY, cls.cpg.Load, backref="loads_from")

    @declared_attr
    def stored_from(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.STORE_MEMORY, cls.cpg.Store, backref="stores_to")

    @declared_attr
    def may_alias(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.MAY_ALIAS, cls, direction=Direction.SYMMETRIC  # type: ignore[arg-type]
        )

    @declared_attr
    def must_alias(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.MUST_ALIAS, cls, direction=Direction.SYMMETRIC  # type: ignore[arg-type]
        )

    @declared_attr
    def contained_by(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.CONTAINS,
            cls,  # type: ignore[arg-type]
            direction=Direction.IN,
            backref="contains",
        )

    @declared_attr
    def subregion_of(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.SUBREGION,
            cls,  # type: ignore[arg-type]
            direction=Direction.IN,
            backref="subregions",
        )

    @declared_attr
    def accessed_by(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Instruction,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=or_(
                and_(
                    cls.cpg.Edge.kind == EdgeKind.LOAD_MEMORY,
                    cls.cpg.Edge.source == cls.uuid,
                ),
                and_(
                    cls.cpg.Edge.kind == EdgeKind.STORE_MEMORY,
                    cls.cpg.Edge.target == cls.uuid,
                ),
            ),
            secondaryjoin=or_(
                and_(
                    cls.cpg.Edge.kind == EdgeKind.LOAD_MEMORY,
                    cls.cpg.Edge.target == cls.cpg.Instruction.uuid,
                ),
                and_(
                    cls.cpg.Edge.kind == EdgeKind.STORE_MEMORY,
                    cls.cpg.Edge.source == cls.cpg.Instruction.uuid,
                ),
            ),
            backref=backref(
                "accesses", doc="Connects an instruction to memory locations that it loads/stores"
            ),
            lazy="dynamic",
            doc="Connects a memory location to instructions that may load or store from the location.",
        )

    @declared_attr
    def allocated_by(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.ALLOCATES,
                cls.cpg.Edge.target == cls.uuid,
            ),
            secondaryjoin=cls.cpg.Edge.source == cls.cpg.Node.uuid,
            lazy="dynamic",
            doc="Connects a memory location to the instruction that allocated it.",
            backref="allocates",
        )

    @declared_attr
    def pointers(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.POINTS_TO,
                cls.cpg.Edge.target == cls.uuid,
            ),
            secondaryjoin=cls.cpg.Node.uuid == cls.cpg.Edge.source,
            backref="points_to",
            lazy="dynamic",
            doc="Connects a memory location to other locations or values that point to it.",
        )


class DataflowSignature(NodeMixin):
    _kind = NodeKind.DATAFLOW_SIGNATURE

    @declared_attr
    def signature_for(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DATAFLOW_SIGNATURE_FOR_CALLSITE,
            cls.cpg.Instruction,
            backref="dataflow_signatures",
        )

    @declared_attr
    def signature_for_function(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DATAFLOW_SIGNATURE_FOR_FUNCTION,
            cls.cpg.Function,
            backref="dataflow_signatures",
        )

    @declared_attr
    def directly_flows_from(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.DIRECT_DATAFLOW_SIGNATURE,
                cls.cpg.Edge.target == cls.uuid,
            ),
            secondaryjoin=cls.cpg.Node.uuid == cls.cpg.Edge.source,
            backref="directly_flows_to_dataflow_signature",
            lazy="dynamic",
            doc="Connects a dataflow signature to all nodes representing data that flows directly into the modeled operation.",
        )

    @declared_attr
    def indirectly_flows_from(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.INDIRECT_DATAFLOW_SIGNATURE,
                cls.cpg.Edge.target == cls.uuid,
            ),
            secondaryjoin=cls.cpg.Node.uuid == cls.cpg.Edge.source,
            backref="indirectly_flows_to_dataflow_signature",
            lazy="dynamic",
            doc="Connects a dataflow signature to all nodes representing data that flows indirectly into the modeled operation.",
        )

    @declared_attr
    def control_flows_from(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.CONTROL_DATAFLOW_SIGNATURE,
                cls.cpg.Edge.target == cls.uuid,
            ),
            secondaryjoin=cls.cpg.Node.uuid == cls.cpg.Edge.source,
            backref="control_flows_to_dataflow_signature",
            lazy="dynamic",
            doc="Connects a dataflow signature to all nodes representing whether data flows into the modeled operation.",
        )

    @declared_attr
    def flows_to(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.DATAFLOW_SIGNATURE, cls.cpg.Edge.source == cls.uuid
            ),
            secondaryjoin=cls.cpg.Node.uuid == cls.cpg.Edge.target,
            backref="flows_from_dataflow_signature",
            lazy="dynamic",
            doc="Connects a dataflow signature to all nodes representing data that flows out of the modeled operation.",
        )

    @hybrid_property
    def is_allocator(self) -> bool:
        return self.deallocator is not None

    @is_allocator.expression  # type: ignore[no-redef]
    def is_allocator(cls) -> bool:
        return cls.deallocator.isnot(None)


class InputSignature(NodeMixin):
    _kind = NodeKind.INPUT_SIGNATURE

    @declared_attr
    def signature_for(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DATAFLOW_SIGNATURE_FOR_CALLSITE,
            cls.cpg.Instruction,
            backref="input_signatures",
        )

    @declared_attr
    def signature_for_function(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DATAFLOW_SIGNATURE_FOR_FUNCTION,
            cls.cpg.Function,
            backref="input_signatures",
        )

    @declared_attr
    def flows_to(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.DATAFLOW_SIGNATURE, cls.cpg.Edge.source == cls.uuid
            ),
            secondaryjoin=cls.cpg.Node.uuid == cls.cpg.Edge.target,
            backref="flows_from_input_signature",
            lazy="dynamic",
            doc="Connects an input signature to data values that may be affected by the modeled operation.",
        )


class OutputSignature(NodeMixin):
    _kind = NodeKind.OUTPUT_SIGNATURE

    @declared_attr
    def signature_for(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DATAFLOW_SIGNATURE_FOR_CALLSITE,
            cls.cpg.Instruction,
            backref="output_signatures",
        )

    @declared_attr
    def signature_for_function(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DATAFLOW_SIGNATURE_FOR_FUNCTION,
            cls.cpg.Function,
            backref="output_signatures",
        )

    @declared_attr
    def flows_from(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.DATAFLOW_SIGNATURE, cls.cpg.Edge.target == cls.uuid
            ),
            secondaryjoin=cls.cpg.Node.uuid == cls.cpg.Edge.source,
            backref="flows_to_output_signature",
            lazy="dynamic",
            doc="Connects an output signature to data values that may be affect the modeled operation.",
        )



"""This module contains helpers for interacting with MATE's object storage service."""

import logging

from minio import Minio
from minio.error import BucketAlreadyExists, BucketAlreadyOwnedByYou

from mate_query.config import MATE_STORAGE_ACCESS_KEY, MATE_STORAGE_SECRET_KEY

logger = logging.getLogger(__name__)

_connection_string = None
client: Minio = None


def initialize(connection_string: str) -> None:
    global client, _connection_string

    if client is not None:
        logger.error(
            f"MATE has already had its storage service initialized with {_connection_string}; skipping"
        )
        return

    _connection_string = connection_string
    client = Minio(
        connection_string,
        access_key=MATE_STORAGE_ACCESS_KEY,
        secret_key=MATE_STORAGE_SECRET_KEY,
        secure=False,
    )

    try:
        client.make_bucket("artifacts")
    except (BucketAlreadyOwnedByYou, BucketAlreadyExists):
        pass



"""A base class for declarative mapping, with extra functionality.

See https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/mixins.html#augmenting-the-base
"""

from __future__ import annotations

import enum
from threading import local
from typing import TYPE_CHECKING, Any, Type, TypeVar

from sqlalchemy import Column
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.hybrid import hybrid_method
from sqlalchemy.orm import sessionmaker
from sqlalchemy.types import Enum as SAEnum
from sqlalchemy.types import String

from mate_query.cpg.models._typechecking import HasDynamicAttributes

T = TypeVar("T", bound="Base")

_VALUES_CALLABLE = lambda x: [e.value for e in x]


def enum_column(name: str, klass: Type[enum.Enum]) -> Column:
    """Create an enum column named ``name`` from the given ``enum.Enum`` subclass."""
    return Column(
        name,
        SAEnum(klass, values_callable=_VALUES_CALLABLE, validate_strings=True, index=True),
        nullable=False,
    )


def uuid_column() -> Column:
    return Column("uuid", String, primary_key=True, index=True)


if TYPE_CHECKING:

    # NOTE(lb): It's unclear to me why simply importing this exact definition from
    # _typechecking will result in a type error ("error: Invalid metaclass
    # 'ClassHasDynamicAttributes'"). Mypy bug?
    class ClassHasDynamicAttributes(type, HasDynamicAttributes):
        pass

else:

    ClassHasDynamicAttributes = type


class Base(HasDynamicAttributes, metaclass=ClassHasDynamicAttributes):
    uuid = uuid_column()
    attributes = Column(JSONB, nullable=False)
    local: Any = None

    @classmethod
    def _session(cls) -> Any:
        return sessionmaker(bind=cls.local.bind)()

    @classmethod
    def find_by(cls: Type[T], **kwargs: Any) -> T:
        return cls._session().query(cls).filter_by(**kwargs).one()

    @classmethod
    def from_uuid(cls: Type[T], uuid: str) -> T:
        return cls.find_by(uuid=uuid)

    @hybrid_method
    def __eq__(self: T, other: T) -> bool:
        return self.uuid == other.uuid

    @hybrid_method
    def __hash__(self) -> int:
        return hash(self.uuid)


Base.local = local()



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Optional, Tuple

from sqlalchemy.dialects.postgresql import ARRAY, array
from sqlalchemy.orm import aliased
from sqlalchemy.sql.expression import CTE, case, func, literal, true, tuple_
from sqlalchemy.types import String

import mate_query.db as db
from mate_query.cfl import CSCFGPath
from mate_query.db import BOT

if TYPE_CHECKING:
    from mate_query.cpg.models.node.ast.llvm import Instruction
    from mate_query.db import Graph as CPG


def control_flow_avoiding(
    cpg: CPG,
    start_cte: CTE,
    avoid_map_cte: CTE,
    stop_map_cte: CTE,
    stack_bound: Optional[int] = None,
    limit: Optional[int] = None,
) -> Iterable[Tuple[Optional[str], Instruction, str, Instruction, str]]:
    """Finds pairs of instruction/contexts from ``start_cte`` and ``stop_map_cte`` that are
    reachable in the control-flow graph without passing through the designated nodes in
    ``avoid_map_cte``.

    start_cte must be a CTE with columns ``info``, ``start_uuid`` and ``start_context``.
    avoid_map_cte must be a CTE with columns ``info``, ``start_uuid``, ``start_context``,
    ``avoid_uuid``, and ``avoid_context``. stop_map_cte must be a CTE with columns ``info``,
    ``start_uuid``, ``start_context``, ``stop_uuid``, and ``stop_context``.
    """
    InitialConfiguration = (
        cpg.session.query(start_cte)
        .with_entities(
            start_cte.c.info.label("info"),
            start_cte.c.uuid.label("uuid"),
            array([start_cte.c.context, BOT]).cast(ARRAY(String)).label("stack"),
        )
        .cte()
    )

    AvoidMap = (
        cpg.session.query(avoid_map_cte)
        .with_entities(
            avoid_map_cte.c.info,
            avoid_map_cte.c.start_context,
            avoid_map_cte.c.start_uuid,
            avoid_map_cte.c.avoid_context,
            avoid_map_cte.c.avoid_uuid,
        )
        .cte()
    )

    StopMap = (
        cpg.session.query(stop_map_cte)
        .with_entities(
            stop_map_cte.c.info,
            stop_map_cte.c.start_context,
            stop_map_cte.c.start_uuid,
            stop_map_cte.c.stop_context,
            stop_map_cte.c.stop_uuid,
        )
        .cte()
    )

    PathQuery = (
        db.PathBuilder(PathBase=CSCFGPath)
        .initial_configuration(InitialConfiguration)
        .continuing_while(
            lambda PDA, Edge: (
                (
                    tuple_(
                        PDA.c.info,
                        func.split_part(PDA.c.source_stack[1], literal("-->"), 2),
                        PDA.c.source,
                        func.split_part(PDA.c.stack_top, literal("-->"), 2),
                        Edge.source,
                    ).notin_(AvoidMap)
                )
                & (
                    tuple_(
                        PDA.c.info,
                        PDA.c.source_stack[1],
                        PDA.c.source,
                        func.split_part(PDA.c.stack_top, literal("-->"), 2),
                        Edge.source,
                    ).notin_(AvoidMap)
                )
                & (
                    tuple_(
                        PDA.c.info,
                        func.split_part(PDA.c.source_stack[1], literal("-->"), 2),
                        PDA.c.source,
                        PDA.c.stack_top,
                        Edge.source,
                    ).notin_(AvoidMap)
                )
                & (
                    tuple_(
                        PDA.c.info,
                        PDA.c.source_stack[1],
                        PDA.c.source,
                        PDA.c.stack_top,
                        Edge.target,
                    ).notin_(AvoidMap)
                )
                & (~(PDA.c.stack[2 : func.cardinality(PDA.c.stack)].any(PDA.c.stack_top)))
                & (
                    (func.cardinality(PDA.c.stack) <= stack_bound)
                    if stack_bound is not None
                    else true()
                )
            )
        )
        .build(cpg)
    )

    StartInst = aliased(cpg.Instruction)
    StopInst = aliased(cpg.Instruction)

    query = (
        cpg.session.query(PathQuery)
        .filter(
            (
                tuple_(
                    PathQuery.info,
                    PathQuery.source_stack[1],
                    PathQuery.source,
                    func.split_part(PathQuery.stack_top, literal("-->"), 2),
                    PathQuery.target,
                ).in_(StopMap)
            )
            | (
                tuple_(
                    PathQuery.info,
                    PathQuery.source_stack[1],
                    PathQuery.source,
                    PathQuery.stack_top,
                    PathQuery.target,
                ).in_(StopMap)
            )
        )
        .join(StartInst, PathQuery.source == StartInst.uuid)
        .join(StopInst, PathQuery.target == StopInst.uuid)
        .with_entities(
            PathQuery.info,
            StartInst,
            PathQuery.source_stack[1],
            StopInst,
            case(
                [
                    (
                        PathQuery.stack_top.contains("-->"),
                        func.split_part(PathQuery.stack_top, literal("-->"), 2),
                    )
                ],
                else_=PathQuery.stack_top,
            ),
        )
        .distinct()
    )

    findings = query.all() if limit is None else query.limit(limit).all()

    return findings



"""This module contains implementation details; it is not part of the MATE API.

Helpers for deriving documentation and relationships from the MATE JSON schemata.
"""
from __future__ import annotations

from collections.abc import Iterable
from enum import Enum
from typing import TYPE_CHECKING

from sqlalchemy import and_
from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.orm import backref as backref_
from sqlalchemy.orm import relationship

from mate_common.assertions import mate_assert
from mate_common.models.cpg_types import (
    EdgeKind,
    LLVMIntrinsic,
    NodeJSON,
    NodeKind,
    Opcode,
    Relationship,
)
from mate_common.schemata import EDGE_SCHEMA_BY_KIND, ENDPOINTS, NODE_SCHEMA_BY_KIND, RELATIONSHIPS
from mate_query.cpg.models.docs import description_for_node_kind
from mate_query.string import StringComparator

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Optional, Type

    from mate_query.cpg.models.core.node import Node
    from mate_query.db import Graph as CPG


class Direction(Enum):
    IN = "in"
    OUT = "out"
    SYMMETRIC = "symmetric"  # For relationships where either IN or OUT would be suitable


class EnumComparator(Comparator):
    """A specialized SQLAlchemy Comparator for ``enum.Enum``."""

    @staticmethod
    def __cast__(value: Any) -> Any:
        """Turns an ``Enum`` or an iterable into its underlying value(s), if appropriate."""
        if isinstance(value, Enum):
            return value.value

        # NOTE(ww): We support direct comparisons to strings and bytes, which happen
        # to be iterable. Check them here to prevent unintended recursion.
        if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
            return [EnumComparator.__cast__(v) for v in value]
        return value

    def operate(self, op: Any, other: Any, **_kwargs: Any) -> Any:
        return op(self.__clause_element__(), EnumComparator.__cast__(other))


# Map node kinds and attribute names to the Python enums that represent their
# values
_ENUMS = {
    (NodeKind.INSTRUCTION, NodeJSON.OPCODE.value): Opcode,
    (NodeKind.INSTRUCTION, NodeJSON.INTRINSIC.value): LLVMIntrinsic,
}

# NOTE(ww): We include "array" and "object" here for completeness, but we don't coerce
# their inner elements (since SQLAlchemy only needs explicit coercions for scalar properties).
_JSON_SCHEMA_TYPE_TO_PYTHON_TYPE = {
    "boolean": bool,
    "integer": int,
    "string": str,
    "array": list,
    "object": dict,
}


def _coerce_attr(attr: Any, type_: Type) -> Any:
    """Given a class-level SQLAlchemy JSON attribute, attempt to call the appropriate type coercion
    method for ``type_``."""
    if type_ == str:
        return attr.as_string()
    elif type_ == bool:
        return attr.as_boolean()
    elif type_ == int:
        return attr.as_integer()

    mate_assert(type_ in (list, dict), f"uncoercable type: {type_}")
    return attr


def _is_many_end(this_is_source: bool, edge_kind: EdgeKind) -> bool:
    """Decide whether or not to use a list at one end of a ``relationship``"""
    this_is_target = not this_is_source
    rel = RELATIONSHIPS[edge_kind]
    if rel == Relationship.ONE_TO_ONE:
        return False
    if (
        rel == Relationship.MANY_TO_ONE
        and this_is_source
        or rel == Relationship.ONE_TO_MANY
        and this_is_target
    ):
        return False
    if (
        rel == Relationship.ONE_TO_MANY
        and this_is_source
        or rel == Relationship.MANY_TO_ONE
        and this_is_target
    ):
        return True
    if rel == Relationship.MANY_TO_MANY:
        return True
    mate_assert(False, f"Unreachable: {edge_kind}")
    return False


def _use_list_arguments(this_is_source: bool, edge_kind: EdgeKind) -> Dict:
    if _is_many_end(this_is_source, edge_kind):
        return {}
    return {"uselist": False, "lazy": "select"}


def _doc_string(this_is_source: bool, edge_kind: EdgeKind) -> Optional[str]:
    return EDGE_SCHEMA_BY_KIND[edge_kind.value].get(
        "description" if this_is_source else "backref_description"
    )


def _backref_doc(this_is_source: bool, edge_kind: EdgeKind) -> Dict:
    desc = _doc_string(this_is_source, edge_kind)
    if desc:
        return {"doc": desc}
    return {}


def make_edge_relationship(
    cpg: CPG,
    edge_kind: EdgeKind,
    this: Type[Node],
    other: Type[Node],
    direction: Optional[Direction] = None,
    back_populates: Optional[str] = None,
    backref: Optional[str] = None,
) -> RelationshipProperty:
    """Generate a SQLAlchemy relationship given an edge kind."""
    summary = f"kind: {edge_kind}, this: {this}, other: {other}"
    if this == other:
        mate_assert(direction is not None, f"Specified direction required on self-edges: {summary}")
        mate_assert(
            0
            < len(
                [
                    spec
                    for spec in ENDPOINTS[edge_kind]
                    if this._kind in spec.sources and this._kind in spec.targets
                ]
            ),
            f"Invalid relationship spec, no match found in schema: {summary}",
        )
        # NOTE(lb): If the Direction is SYMMETRIC, we can make an arbitrary choice.
        this_is_source = direction == Direction.OUT
    else:
        # Deduce the direction from the spec
        candidates = [
            spec
            for spec in ENDPOINTS[edge_kind]
            if (this._kind in spec.sources and other._kind in spec.targets)
            or (this._kind in spec.targets and other._kind in spec.sources)
        ]
        mate_assert(
            len(candidates) > 0,
            f"Invalid edge relationship spec: {summary}, endpoints: {ENDPOINTS[edge_kind]}",
        )
        mate_assert(
            len(candidates) == 1,
            f"Ambiguous edge relationship spec: {summary}, {ENDPOINTS[edge_kind]}",
        )
        spec = candidates[0]
        this_is_source = this._kind in spec.sources

    # Decide whether or not to use a list (at either end)
    extra_args = _use_list_arguments(this_is_source, edge_kind)
    if back_populates is not None:
        extra_args["back_populates"] = back_populates
    if backref is not None:
        extra_args["backref"] = backref_(
            backref,
            **_use_list_arguments(not this_is_source, edge_kind),
            **_backref_doc(not this_is_source, edge_kind),
        )

    # Figure out what ends of the edge the models should be on
    this_edge_end = cpg.Edge.source if this_is_source else cpg.Edge.target
    other_edge_end = cpg.Edge.target if this_is_source else cpg.Edge.source

    return relationship(
        other,
        secondary=cpg.Edge.__table__,
        primaryjoin=and_(
            this.uuid == this_edge_end,
            cpg.Edge.kind == edge_kind,
        ),
        secondaryjoin=and_(
            other.uuid == other_edge_end,
        ),
        doc=_doc_string(this_is_source, edge_kind),
        **extra_args,
    )


def _make_hybrid_property(
    name: str, type_: Type, enum: Optional[Type[Enum]] = None, docs: Optional[str] = None
) -> Callable:
    @hybrid_property
    def getter0(self: Any) -> Any:
        attr = self.attributes[name]
        if enum is not None:
            return enum(attr)
        return attr

    # If we're building an expression for a field that has a mapped enum,
    # we construct a comparator instead of an expression. This allows
    # us to do e.g. `filter_by(prop=enum)` instead of `filter_by(prop=enum.value)`.
    #
    # Additionally, if the field in question is a string, then we use a custom
    # comparator with regex support (for the purposes of signatures).
    if enum is None:

        if type_ == str:

            @getter0.comparator
            def getter1(cls: Type[Node]) -> Any:
                attr = cls.attributes[name]
                return StringComparator(_coerce_attr(attr, type_))

        else:

            @getter0.expression
            def getter1(cls: Type[Node]) -> Any:
                attr = cls.attributes[name]
                return _coerce_attr(attr, type_)

    else:

        @getter0.comparator
        def getter1(cls: Type[Node]) -> Any:
            attr = cls.attributes[name]
            return EnumComparator(_coerce_attr(attr, type_))

    # NOTE(ww): Cheese Sphinx into adding the return type to these properties.
    # TODO(ww): Why doesn't this work?
    # getter0.fget.__annotations__ = {"return": type_}
    # getter1.fget.__annotations__ = {"return": type_}
    if docs is None:
        docs = f":rtype: {type_.__name__}"
    else:
        docs = f"{docs}\n\n\t:rtype: {type_.__name__}"

    getter0.__name__ = name
    getter1.__name__ = name
    getter0.__doc__ = docs
    getter1.__doc__ = docs
    return getter1


def _hybrid_properties_for_kind(cls: Type[Node]) -> Dict[str, hybrid_property]:
    properties: Dict[str, hybrid_property] = dict()
    # Attach hybrid properties
    for (property_name, property_schema) in NODE_SCHEMA_BY_KIND[cls._kind.value][
        "properties"
    ].items():
        # Don't overwrite properties that were customized/already set
        if property_name == "node_kind" or getattr(cls, property_name, None) is not None:
            continue

        # Types can either be specified directly, or we can deduce the
        # type from one-element "enum" clauses
        if "type" in property_schema:
            schema_type = property_schema["type"]
            mate_assert(
                schema_type in _JSON_SCHEMA_TYPE_TO_PYTHON_TYPE,
                f"Unmapped JSON schema type: {schema_type}",
            )
            type_: Type = _JSON_SCHEMA_TYPE_TO_PYTHON_TYPE[schema_type]
        elif "enum" in property_schema:
            enum = property_schema["enum"]
            mate_assert(len(enum) == 1, f"Non-unary enum with no explicit type: {enum}")
            type_ = type(enum[0])
        else:
            message = (
                f"Schema error: can't infer Python type for {property_name}: {property_schema}"
            )
            if "$ref" in property_schema:
                message = (
                    f"{message}\nHint: {property_name} looks like it contains a reference; "
                    "maybe increase the gas limit in schemata._organize_schema?"
                )
            mate_assert(False, message)

        properties[property_name] = _make_hybrid_property(
            property_name,
            type_=type_,
            enum=_ENUMS.get((cls._kind, property_name)),
            docs=property_schema.get("description"),
        )
    return properties


def add_properties_to_node_class(cls: Type[Node]) -> Any:
    """Add documentation and hybrid properties from the JSON schema to a node class."""
    cls.__doc__ = description_for_node_kind(cls._kind.value)
    for (name, prop) in _hybrid_properties_for_kind(cls).items():
        setattr(cls, name, prop)



from typing import Any, List


def info(obj_list: List[Any]) -> None:
    """Shorthand to display useful information given a list of Nodes or Edges."""
    for x in obj_list:
        print(x, x.attributes)



"""This module exports a class that can be used as a base class for any class that is used in such a
way that Mypy can't figure out if/how it's attributes are defined. The two canonical examples are:

- Classes with a custom implementation of ``__getattr__``.
- Subclasses of Node, which have many attributes that are set dynamically by
  ``setattr``.

See https://mypy.readthedocs.io/en/latest/cheat_sheet_py3.html.
"""

from typing import TYPE_CHECKING, Any


class _HasDynamicAttributes:
    def __getattr__(self, name: str) -> Any:
        raise AttributeError(name)


class _ClassHasDynamicAttributes(type, _HasDynamicAttributes):
    """This is a metaclass version of ``HasDynamicAttributes``.

    It's effect is to convince Mypy that a class has dynamically set *class variables* rather than
    dynamic *instance variables*.
    """

    pass


if TYPE_CHECKING:

    HasDynamicAttributes = _HasDynamicAttributes
    ClassHasDynamicAttributes = _ClassHasDynamicAttributes

else:

    HasDynamicAttributes = object
    ClassHasDynamicAttributes = type



"""Retrieve documentation for the CPG models from the JSON schemata."""

import logging
from typing import Any, Dict

from mate_common.schemata import NODE_SCHEMA_BY_KIND

logger = logging.getLogger(__name__)

_NO_DOCS = "No documentation available"


def _no_docs(message: str) -> str:
    """Gracefully put a message where documentation should be."""
    return f"{_NO_DOCS}: {message}"


def _description_for_kind(schema_by_kind: Dict[str, Any], kind: str) -> str:
    """Retrieve the JSON schema 'description' field for this node/edge kind."""
    try:
        if schema_by_kind is not None:
            return schema_by_kind[kind]["description"]
    except KeyError:
        logger.warning(f"Malformed JSON schema for {kind}")
    return _no_docs(f"Malformed JSON schema for {kind}")


def description_for_node_kind(kind: str) -> str:
    return _description_for_kind(NODE_SCHEMA_BY_KIND, kind)



# NOTE(lb) With the resolution of #1044, the majority of this module will
# disappear.
from __future__ import annotations

from typing import List, Optional, Type

from mate_common.assertions import mate_assert

# TODO(ww): Something about Type[...] really screws with pylint here.
from mate_query.cpg.models.core.edge import Edge  # pylint: disable=unused-import
from mate_query.cpg.models.core.node import Node  # pylint: disable=unused-import
from mate_query.cpg.models.core.node import Node as BaseNode
from mate_query.cpg.models.node.analysis import (
    DataflowSignature,
    InputSignature,
    MemoryLocation,
    OutputSignature,
)
from mate_query.cpg.models.node.ast.bin import ASMBlock, ASMGlobalVariable, ASMInst, PLTStub, VTable
from mate_query.cpg.models.node.ast.llvm import (
    Alloca,
    Argument,
    Block,
    Call,
    CallReturn,
    CallSite,
    Constant,
    ConstantFP,
    ConstantInt,
    ConstantString,
    ConstantUndef,
    Function,
    GlobalVariable,
    Instruction,
    Intrinsic,
    Invoke,
    LLVMType,
    Load,
    LocalVariable,
    Memcpy,
    Memset,
    Module,
    ParamBinding,
    Resume,
    Ret,
    Store,
    UnclassifiedNode,
    Variable,
)
from mate_query.cpg.models.node.ast.mc import MachineBasicBlock, MachineFunction, MachineInstr
from mate_query.cpg.models.node.dwarf import (
    ArrayType,
    BasicType,
    ClassType,
    CompositeCachedType,
    CompositeType,
    DerivedType,
    DWARFArgument,
    DWARFLocalVariable,
    DWARFType,
    EnumType,
    StructureType,
    SubroutineType,
    UnionType,
)
from mate_query.cpg.models.node.translation_unit import TranslationUnit


# TODO(1044): Delete this class, merge into Postgres CPG class
class BaseCPG:
    Node: Type[Node]
    Edge: Type[Edge]

    ASMBlock: Type[ASMBlock]
    ASMGlobalVariable: Type[ASMGlobalVariable]
    ASMInst: Type[ASMInst]
    Alloca: Type[Alloca]
    Argument: Type[Argument]
    ArrayType: Type[ArrayType]
    BasicType: Type[BasicType]
    Block: Type[Block]
    Call: Type[Call]
    CallReturn: Type[CallReturn]
    CallSite: Type[CallSite]
    ClassType: Type[ClassType]
    CompositeCachedType: Type[CompositeCachedType]
    CompositeType: Type[CompositeType]
    Constant: Type[Constant]
    ConstantFP: Type[ConstantFP]
    ConstantInt: Type[ConstantInt]
    ConstantString: Type[ConstantString]
    ConstantUndef: Type[ConstantUndef]
    DWARFType: Type[DWARFType]
    DWARFArgument: Type[DWARFArgument]
    DWARFLocalVariable: Type[DWARFLocalVariable]
    DataflowSignature: Type[DataflowSignature]
    DerivedType: Type[DerivedType]
    EnumType: Type[EnumType]
    Function: Type[Function]
    GlobalVariable: Type[GlobalVariable]
    LocalVariable: Type[LocalVariable]
    InputSignature: Type[InputSignature]
    Instruction: Type[Instruction]
    Intrinsic: Type[Intrinsic]
    Invoke: Type[Invoke]
    LLVMType: Type[LLVMType]
    Load: Type[Load]
    MachineBasicBlock: Type[MachineBasicBlock]
    MachineFunction: Type[MachineFunction]
    MachineInstr: Type[MachineInstr]
    Memcpy: Type[Memcpy]
    MemoryLocation: Type[MemoryLocation]
    Memset: Type[Memset]
    Module: Type[Module]
    OutputSignature: Type[OutputSignature]
    ParamBinding: Type[ParamBinding]
    PLTStub: Type[PLTStub]
    Resume: Type[Resume]
    Ret: Type[Ret]
    Store: Type[Store]
    StructureType: Type[StructureType]
    SubroutineType: Type[SubroutineType]
    TranslationUnit: Type[TranslationUnit]
    UnclassifiedNode: Type[UnclassifiedNode]
    UnionType: Type[UnionType]
    Variable: Type[Variable]
    VTable: Type[VTable]

    _node_model_classes: List[Type[BaseNode]]

    def _attach_node_models(self, name: str) -> None:
        mate_assert(self.Node is not None)

        def make_node_class(
            cpg_parent: Type[BaseNode], model_parent: Type, class_name: Optional[str] = None
        ) -> Type:
            mate_assert(cpg_parent != model_parent)
            if class_name is None:
                class_name = model_parent.__name__

            return type(
                class_name,
                (model_parent, cpg_parent),
                {"cpg": self, "build_id": name},
            )

        # NOTE(lb): These classes are in reverse relationship dependency order,
        # i.e. if B has a relationship that mentions A, then A comes before B.
        #
        # Other than that, they're grouped by parent module and inheritance
        # hierarchy.
        self.TranslationUnit = make_node_class(self.Node, TranslationUnit)
        self.DWARFArgument = make_node_class(self.Node, DWARFArgument)
        self.DWARFLocalVariable = make_node_class(self.Node, DWARFLocalVariable)
        self.DWARFType = make_node_class(self.Node, DWARFType)
        self.ArrayType = make_node_class(self.DWARFType, ArrayType)
        self.BasicType = make_node_class(self.DWARFType, BasicType)
        self.ClassType = make_node_class(self.DWARFType, ClassType)
        self.CompositeCachedType = make_node_class(self.DWARFType, CompositeCachedType)
        self.CompositeType = make_node_class(self.DWARFType, CompositeType)
        self.DerivedType = make_node_class(self.DWARFType, DerivedType)
        self.EnumType = make_node_class(self.DWARFType, EnumType)
        self.StructureType = make_node_class(self.DWARFType, StructureType)
        self.SubroutineType = make_node_class(self.DWARFType, SubroutineType)
        self.UnionType = make_node_class(self.DWARFType, UnionType)

        self.ASMInst = make_node_class(self.Node, ASMInst)
        self.ASMBlock = make_node_class(self.Node, ASMBlock)
        self.ASMGlobalVariable = make_node_class(self.Node, ASMGlobalVariable)
        self.VTable = make_node_class(self.Node, VTable)
        self.PLTStub = make_node_class(self.Node, PLTStub)

        self.MachineInstr = make_node_class(self.Node, MachineInstr)
        self.MachineBasicBlock = make_node_class(self.Node, MachineBasicBlock)
        self.MachineFunction = make_node_class(self.Node, MachineFunction)

        self.Constant = make_node_class(self.Node, Constant)
        self.ConstantFP = make_node_class(self.Constant, ConstantFP)
        self.ConstantInt = make_node_class(self.Constant, ConstantInt)
        self.ConstantString = make_node_class(self.Constant, ConstantString)
        self.ConstantUndef = make_node_class(self.Constant, ConstantUndef)

        self.Variable = make_node_class(self.Node, Variable)
        self.Instruction = make_node_class(self.Node, Instruction)
        # NOTE(lb): CallSite and Intrinsic need to pass class_name explicitly
        # because they don't have _kind attributes (they are "union" classes,
        # they don't have a corresponding node kind).
        self.CallSite = make_node_class(self.Instruction, CallSite, class_name="CallSite")
        self.Call = make_node_class(self.CallSite, Call)
        self.Invoke = make_node_class(self.CallSite, Invoke)
        self.Intrinsic = make_node_class(self.Call, Intrinsic, class_name="Instrinsic")
        self.Memcpy = make_node_class(self.Intrinsic, Memcpy)
        self.Memset = make_node_class(self.Intrinsic, Memset)
        self.Alloca = make_node_class(self.Instruction, Alloca)
        self.Load = make_node_class(self.Instruction, Load)
        self.Resume = make_node_class(self.Instruction, Resume)
        self.Ret = make_node_class(self.Instruction, Ret)
        self.Store = make_node_class(self.Instruction, Store)

        self.Module = make_node_class(self.Node, Module)
        self.Argument = make_node_class(self.Variable, Argument)
        self.Block = make_node_class(self.Node, Block)
        self.Function = make_node_class(self.Node, Function)
        self.GlobalVariable = make_node_class(self.Variable, GlobalVariable)
        self.LLVMType = make_node_class(self.Node, LLVMType)
        self.LocalVariable = make_node_class(self.Variable, LocalVariable)
        self.MemoryLocation = make_node_class(self.Node, MemoryLocation)
        self.DataflowSignature = make_node_class(self.Node, DataflowSignature)
        self.InputSignature = make_node_class(self.Node, InputSignature)
        self.OutputSignature = make_node_class(self.Node, OutputSignature)
        self.UnclassifiedNode = make_node_class(self.Node, UnclassifiedNode)
        self.ParamBinding = make_node_class(self.Node, ParamBinding)
        self.CallReturn = make_node_class(self.Node, CallReturn)

        # These are the groups of interesting inheritance hierarchies, with the
        # root at the top of each subtree.

        self._node_model_classes = [
            self.ASMBlock,
            self.ASMGlobalVariable,
            self.ASMInst,
            self.Variable,
            self.Alloca,
            self.Argument,
            self.ArrayType,
            self.BasicType,
            self.Block,
            self.Call,
            self.CallReturn,
            self.CallSite,
            self.ClassType,
            self.CompositeCachedType,
            self.CompositeType,
            self.Constant,
            self.ConstantFP,
            self.ConstantInt,
            self.ConstantString,
            self.ConstantUndef,
            self.DWARFType,
            self.DWARFArgument,
            self.DWARFLocalVariable,
            self.DataflowSignature,
            self.DerivedType,
            self.EnumType,
            self.Function,
            self.GlobalVariable,
            self.LocalVariable,
            self.InputSignature,
            self.Instruction,
            self.Intrinsic,
            self.Invoke,
            self.LLVMType,
            self.Load,
            self.MachineBasicBlock,
            self.MachineFunction,
            self.MachineInstr,
            self.Memcpy,
            self.MemoryLocation,
            self.Memset,
            self.Module,
            self.OutputSignature,
            self.ParamBinding,
            self.PLTStub,
            self.Resume,
            self.Ret,
            self.Store,
            self.StructureType,
            self.SubroutineType,
            self.TranslationUnit,
            self.UnclassifiedNode,
            self.UnionType,
            self.VTable,
        ]



"""Nodes in the LLVM middle-end's AST.

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""

from __future__ import annotations

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import RelationshipProperty

from mate_common.models.cpg_types import EdgeKind, MachineFrameInfo, NodeKind
from mate_query.cpg.models.node._typechecking import NodeMixin


class MachineFunction(NodeMixin):
    _kind = NodeKind.MACHINE_FUNCTION

    @declared_attr
    def blocks(cls) -> RelationshipProperty:
        """The function's `MachineBasicBlock` list."""
        return cls.edge_relationship(
            EdgeKind.BLOCK_TO_PARENT_FUNCTION,
            cls.cpg.MachineBasicBlock,
            backref="parent_function",
        )

    @declared_attr
    def arguments(cls) -> RelationshipProperty:
        """The arguments of this function."""
        return cls.edge_relationship(
            EdgeKind.MI_FUNCTION_TO_DWARF_ARGUMENT,
            cls.cpg.DWARFArgument,
            backref="machine_function",
        )

    @declared_attr
    def local_variables(cls) -> RelationshipProperty:
        """The local variables in this function."""
        return cls.edge_relationship(
            EdgeKind.MI_FUNCTION_TO_DWARF_LOCAL_VARIABLE,
            cls.cpg.DWARFLocalVariable,
            backref="machine_function",
        )

    @hybrid_property
    def frame(self) -> MachineFrameInfo:
        """A `MachineFrameInfo` model of the function's stack frame."""
        return MachineFrameInfo(**self.frame_info)

    @declared_attr
    def vtables(cls) -> RelationshipProperty:
        """The function's `mate_query.cpg.models.node.ast.bin.VTable` list."""
        return cls.edge_relationship(
            EdgeKind.MI_FUNCTION_TO_VTABLE,
            cls.cpg.VTable,
            backref="machine_functions",
        )

    @declared_attr
    def dwarf_type(cls) -> RelationshipProperty:
        """The function's `mate_query.cpg.models.node.dwarf.DWARFType`."""
        return cls.edge_relationship(
            EdgeKind.HAS_DWARF_TYPE, cls.cpg.DWARFType, backref="machine_functions"
        )


class MachineBasicBlock(NodeMixin):
    _kind = NodeKind.MACHINE_BASIC_BLOCK

    @declared_attr
    def asm_block(cls) -> RelationshipProperty:
        """The corresponding `ASMBlock`."""
        return cls.edge_relationship(
            EdgeKind.MI_BLOCK_TO_ASM_BLOCK, cls.cpg.ASMBlock, backref="mi_block"
        )

    @declared_attr
    def instructions(cls) -> RelationshipProperty:
        """A sequence of `MachineInstr` for each instruction in this machine basic block."""
        return cls.edge_relationship(
            EdgeKind.INSTRUCTION_TO_PARENT_BLOCK, cls.cpg.MachineInstr, backref="mi_block"
        )

    @declared_attr
    def entry(cls) -> RelationshipProperty:
        """The entry `MachineInstr` for this machine basic block."""
        return cls.edge_relationship(EdgeKind.BLOCK_TO_ENTRY_INSTRUCTION, cls.cpg.MachineInstr)

    @declared_attr
    def terminator(cls) -> RelationshipProperty:
        """The exit `MachineInstr` for this machine basic block."""
        return cls.edge_relationship(EdgeKind.BLOCK_TO_TERMINATOR_INSTRUCTION, cls.cpg.MachineInstr)


class MachineInstr(NodeMixin):
    _kind = NodeKind.MACHINE_INSTR



"""Nodes in the source-level AST (as recovered from DWARF debug info)

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""

from __future__ import annotations

from collections import deque
from typing import Any, Dict, List, Optional

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import RelationshipProperty

from mate_common.models.cpg_types import (
    CountSubrange,
    DWARFEnumerator,
    DWARFScope,
    DWARFSubrange,
    DWARFSubrangeKind,
    DWARFTypeCommonInfo,
    DWARFTypeKind,
    EdgeKind,
    GlobalVariableSubrange,
    LocalVariableSubrange,
    NodeKind,
    mate_assert,
)
from mate_query.cpg.models.core.relationships import Direction, EnumComparator
from mate_query.cpg.models.node._typechecking import NodeMixin


class DWARFType(NodeMixin):
    _kind = NodeKind.DWARF_TYPE

    @declared_attr
    def arguments(cls) -> RelationshipProperty:
        """A list of `DWARFArgument` instances that have this type."""
        return cls.edge_relationship(
            EdgeKind.HAS_DWARF_TYPE, cls.cpg.DWARFArgument, backref="dwarf_type"
        )

    @declared_attr
    def local_variables(cls) -> RelationshipProperty:
        """A list of `DWARFLocalVariable` instances that have this type."""
        return cls.edge_relationship(
            EdgeKind.HAS_DWARF_TYPE, cls.cpg.DWARFLocalVariable, backref="dwarf_type"
        )

    @declared_attr
    def parents(cls) -> RelationshipProperty:
        """The parent types of this type."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_PARENT_TYPE, cls, Direction.OUT  # type: ignore[arg-type]
        )

    @declared_attr
    def children(cls) -> RelationshipProperty:
        """The child types of this type."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_PARENT_TYPE, cls, Direction.IN  # type: ignore[arg-type]
        )

    @declared_attr
    def deriving_types(cls) -> RelationshipProperty:
        """Any types that derive directly from this type.

        For the fully expanded derivation set, use `all_derived_types`.
        """
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls, Direction.IN  # type: ignore[arg-type]
        )

    @hybrid_property
    def common(self) -> DWARFTypeCommonInfo:
        """A `DWARFTypeCommonInfo` model of this type's baseline information."""
        return DWARFTypeCommonInfo(**self.dwarf_type["common"])

    @hybrid_property
    def name(self) -> str:
        """The type's name."""
        return self.common.name

    @name.expression  # type: ignore[no-redef]
    def name(cls):
        return cls.dwarf_type["common"]["name"].as_string()

    @hybrid_property
    def tag(self) -> str:
        """The type's tag (i.e., ``DW_TAG_*``)."""
        return self.common.tag

    @tag.expression  # type: ignore[no-redef]
    def tag(cls):
        return cls.dwarf_type["common"]["tag"].as_string()

    @hybrid_property
    def type_kind(self) -> DWARFTypeKind:
        """The type's `DWARFTypeKind`"""
        return DWARFTypeKind(self.dwarf_type["kind"])

    @type_kind.expression  # type: ignore[no-redef]
    def type_kind(cls):
        return EnumComparator(cls.dwarf_type["kind"].as_string())

    # TODO(ww): These predicates can probably be removed now that we can query
    # for explicit DWARFType subclasses.

    @hybrid_property
    def is_void(self) -> bool:
        """Whether this type is a ``void`` type."""
        return self.name == "void" and self.tag == "DW_TAG_unspecified_type"

    @is_void.expression  # type: ignore[no-redef]
    def is_void(cls) -> bool:
        return cls.name == "void" and cls.tag == "DW_TAG_unspecified_type"

    @hybrid_property
    def is_varargs(self) -> bool:
        """Whether this type is a variable argument (vaarg) type."""
        return self.name == "..." and self.tag == "DW_TAG_unspecified_parameters"

    @is_varargs.expression  # type: ignore[no-redef]
    def is_varargs(cls) -> bool:
        return cls.name == "..." and cls.tag == "DW_TAG_unspecified_parameters"

    @hybrid_property
    def is_pointer(self) -> bool:
        """Whether this type is a pointer type."""
        return self.tag == "DW_TAG_pointer_type"

    @is_pointer.expression  # type: ignore[no-redef]
    def is_pointer(cls) -> bool:
        return cls.tag == "DW_TAG_pointer_type"

    @hybrid_property
    def is_member(self) -> bool:
        """Whether this type is a member type."""
        return self.tag == "DW_TAG_member"

    @is_member.expression  # type: ignore[no-redef]
    def is_member(cls) -> bool:
        return cls.tag == "DW_TAG_member"

    @hybrid_property
    def is_const(self) -> bool:
        """Whether this type is ``const``."""
        return self.tag == "DW_TAG_const_type"

    @is_const.expression  # type: ignore[no-redef]
    def is_const(cls) -> bool:
        return cls.tag == "DW_TAG_const_type"

    @hybrid_property
    def is_reference(self) -> bool:
        """Whether this type is an (lvalue) reference."""
        return self.tag == "DW_TAG_reference_type"

    @is_reference.expression  # type: ignore[no-redef]
    def is_reference(cls) -> bool:
        return cls.tag == "DW_TAG_reference_type"

    @hybrid_property
    def is_rvalue_reference(self) -> bool:
        """Whether this type is an rvalue reference."""
        return self.tag == "DW_TAG_rvalue_reference_type"

    @is_rvalue_reference.expression  # type: ignore[no-redef]
    def is_rvalue_reference(cls) -> bool:
        return cls.tag == "DW_TAG_rvalue_reference_type"

    @hybrid_property
    def is_basic(self) -> bool:
        """Whether this type is basic, i.e. `DWARFTypeKind.BASIC`."""
        return self.type_kind == DWARFTypeKind.BASIC

    @hybrid_property
    def is_composite(self) -> bool:
        return self.type_kind == DWARFTypeKind.COMPOSITE

    @hybrid_property
    def is_structure(self) -> bool:
        """Whether this type is a ``struct`` type."""
        return self.type_kind == DWARFTypeKind.STRUCTURE

    @hybrid_property
    def is_array(self) -> bool:
        """Whether this type is an array type."""
        return self.type_kind == DWARFTypeKind.ARRAY

    @hybrid_property
    def is_enum(self) -> bool:
        """Whether this type is an ``enum`` type."""
        return self.type_kind == DWARFTypeKind.ENUM

    @hybrid_property
    def is_union(self) -> bool:
        """Whether this type is a ``union`` type."""
        return self.type_kind == DWARFTypeKind.UNION

    @hybrid_property
    def is_class(self) -> bool:
        """Whether this type is a ``class`` type."""
        return self.type_kind == DWARFTypeKind.CLASS

    @hybrid_property
    def is_derived(self) -> bool:
        """Whether this type is derived from another type."""
        return self.type_kind == DWARFTypeKind.DERIVED

    @hybrid_property
    def is_subroutine(self) -> bool:
        """Whether this type is a subroutine (i.e., function) type."""
        return self.type_kind == DWARFTypeKind.SUBROUTINE

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        try:
            return {"polymorphic_identity": cls._kind}
        except AttributeError:
            return {"polymorphic_identity": NodeKind.DWARF_TYPE}

    @hybrid_property
    def all_derived_types(self) -> List[DWARFType]:
        """Returns a list of all `DWARFType` instances that derive from this type, including
        indirect derivations (e.g. ``T -> T* -> const T*``)."""
        curr = deque(self.deriving_types)
        derived = set()
        while curr:
            base_type = curr.popleft()
            derived.add(base_type)
            curr.extend(base_type.deriving_types)
        return list(derived)


class BasicType(DWARFType):
    _kind = NodeKind.BASIC_TYPE

    @hybrid_property
    def is_unsigned(self) -> bool:
        """Whether this type is unsigned."""
        return self.dwarf_type["unsigned"]


class CompositeType(DWARFType):
    _kind = NodeKind.COMPOSITE_TYPE

    @declared_attr
    def base_type(cls) -> RelationshipProperty:
        """The base type."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @declared_attr
    def elements(cls) -> RelationshipProperty:
        """The member elements of this composite type."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_MEMBER_TYPE, cls.cpg.DWARFType, Direction.OUT
        )


class CompositeCachedType(DWARFType):
    _kind = NodeKind.COMPOSITE_CACHED_TYPE

    @declared_attr
    def recursive_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_RECURSIVE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )


class StructureType(DWARFType):
    _kind = NodeKind.STRUCTURE_TYPE

    @declared_attr
    def base_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @declared_attr
    def members(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_MEMBER_TYPE, cls.cpg.DWARFType, Direction.OUT
        )


class ArrayType(DWARFType):
    _kind = NodeKind.ARRAY_TYPE

    @declared_attr
    def base_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @hybrid_property
    def subrange(self) -> Optional[DWARFSubrange]:
        """Returns a `DWARFSubrange` for this array type's subrange.

        The subrange can be a `CountSubrange`, `GlobalVariableSubrange`, or `LocalVariableSubrange`.
        """
        subrange = self.dwarf_type["subrange"]
        if DWARFSubrangeKind.COUNT.value in subrange:
            return CountSubrange(kind=DWARFSubrangeKind.COUNT, **subrange)
        elif DWARFSubrangeKind.GLOBAL_VARIABLE.value in subrange:
            return GlobalVariableSubrange(
                kind=DWARFSubrangeKind.GLOBAL_VARIABLE,
                **subrange[DWARFSubrangeKind.GLOBAL_VARIABLE.value],
            )
        elif DWARFSubrangeKind.LOCAL_VARIABLE.value in subrange:
            return LocalVariableSubrange(
                kind=DWARFSubrangeKind.LOCAL_VARIABLE,
                **subrange[DWARFSubrangeKind.LOCAL_VARIABLE.value],
            )
        else:
            mate_assert(False, f"Impossible subrange type {subrange}")
        return None


class EnumType(DWARFType):
    _kind = NodeKind.ENUM_TYPE

    @declared_attr
    def base_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @hybrid_property
    def enumerators(self) -> List[DWARFEnumerator]:
        """Returns a list of `DWARFEnumerator` variants."""
        return [DWARFEnumerator(**enum) for enum in self.dwarf_type["enumerators"]]


class UnionType(DWARFType):
    _kind = NodeKind.UNION_TYPE

    @declared_attr
    def base_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @declared_attr
    def members(cls) -> RelationshipProperty:
        """Returns the member types of this ``union``."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_MEMBER_TYPE, cls.cpg.DWARFType, Direction.OUT
        )


class ClassType(DWARFType):
    _kind = NodeKind.CLASS_TYPE

    @declared_attr
    def base_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @declared_attr
    def members(cls) -> RelationshipProperty:
        """Returns the member types of this ``class``."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_MEMBER_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @declared_attr
    def template_params(cls) -> RelationshipProperty:
        """Returns the template parameter types for this ``class``."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_TEMPLATE_PARAM_TYPE, cls.cpg.DWARFType, Direction.OUT
        )


class DerivedType(DWARFType):
    _kind = NodeKind.DERIVED_TYPE

    @declared_attr
    def base_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_BASE_TYPE, cls.cpg.DWARFType, Direction.OUT
        )


class SubroutineType(DWARFType):
    _kind = NodeKind.SUBROUTINE_TYPE

    @declared_attr
    def return_type(cls) -> RelationshipProperty:
        """Returns the subroutine's return type."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_RETURN_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @declared_attr
    def param_types(cls) -> RelationshipProperty:
        """Returns the parameter type for each parameter in the subroutine."""
        return cls.edge_relationship(
            EdgeKind.DWARF_TYPE_TO_PARAM_TYPE, cls.cpg.DWARFType, Direction.OUT
        )

    @hybrid_property
    def has_varargs(self) -> bool:
        """Returns whether the subroutine has a variable number of arguments."""
        return any(param_type.is_varargs for param_type in self.param_types)


class DWARFLocalVariable(NodeMixin):
    """Represents a DWARF view of a local variable."""

    _kind = NodeKind.DWARF_LOCAL_VARIABLE

    @property
    def dwarf_scope(self) -> DWARFScope:
        """Returns the `DWARFScope` for this local variable."""
        return DWARFScope(**self.attributes["dwarf_scope"])


class DWARFArgument(NodeMixin):
    """Represents a DWARF view of a formal function parameter."""

    _kind = NodeKind.DWARF_ARGUMENT

    @property
    def dwarf_scope(self) -> DWARFScope:
        """Returns the `DWARFScope` for this function argument."""
        return DWARFScope(**self.attributes["dwarf_scope"])



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from smt2lib.SMTLIBv2Parser import SMTLIBv2Parser

from mate_common.models.integration import (
    Addr,
    Assertion,
    Constraint,
    Register,
    SMTIdentifier,
    Waypoint,
)

if TYPE_CHECKING:
    from typing import Any, Iterable, List, Optional

    from sqlalchemy.orm import Session

    from mate_query.cpg.models import Edge, Node
    from mate_query.cpg.models.node.ast.bin import ASMBlock
    from mate_query.cpg.models.node.ast.llvm import Instruction
    from mate_query.db import Graph as CPG
    from mate_query.db import Path


# Replacement token recognized by Manticore
# Remove the surrounding ' characters
SMT_REPLACE_TOK = SMTLIBv2Parser.literalNames[SMTLIBv2Parser.Replace].strip("'")


def _construct_waypoint_for_blocks(
    asm_block: ASMBlock, next_asm_block: Optional[ASMBlock]
) -> Waypoint:
    """Construct a waypoint that should be active during the last instruction in ``asm_block``,
    where ``next_asm_block`` is where control flows next."""
    # TODO: Need mypy support for protobuf to use correct Assertion type
    asserts: List[Any] = []
    va = Addr(va=asm_block.terminator.va)
    if next_asm_block is not None and (
        next_asm_block.mi_block.ir_block.parent_function
        == asm_block.mi_block.ir_block.parent_function
    ):
        # If this is an intra-procedural jump, we know that we are jumping
        # to the beginning of the next block, and constrain the instruction
        # pointer accordingly.
        asserts = [
            Assertion(
                location=va,
                constraint=[
                    Constraint(
                        id=[SMTIdentifier(identifier=Register.RIP)],
                        expr=f"(assert (= {SMT_REPLACE_TOK}0 {next_asm_block.va}))",
                    )
                ],
            )
        ]
    return Waypoint(start=va, end=va, asserts=asserts, replacements=[])


def cfg_path_to_waypoints(cpg: CPG, session: Session, cfg_path: Path) -> Iterable[Waypoint]:
    (instrs1, instrs2) = itertools.tee(cfg_path_to_instructions(cpg, session, cfg_path))
    next(instrs2)
    for llvm_instr, next_llvm_instr in itertools.zip_longest(instrs1, instrs2):
        llvm_block = llvm_instr.parent_block

        if llvm_instr == llvm_block.terminator:
            mi_blocks = llvm_block.sorted_mi_blocks()
            if len(mi_blocks) <= 0:
                continue
            mi_block = mi_blocks[-1]

            # Skip fall-through blocks, except the last block/instruction on the path
            if mi_block.can_fallthrough and next_llvm_instr is not None:
                continue

            asm_block = mi_block.asm_block

            va = Addr(va=asm_block.terminator.va)
            no_assert = Waypoint(start=va, end=va, asserts=[], replacements=[])
            if next_llvm_instr is None:  # this was the last instruction
                yield no_assert
                continue

            next_mi_blocks = next_llvm_instr.parent_block.sorted_mi_blocks()
            if len(next_mi_blocks) > 0:
                yield _construct_waypoint_for_blocks(asm_block, next_mi_blocks[0].asm_block)
            else:
                yield no_assert

        elif llvm_instr == llvm_block.entry:
            mi_blocks = llvm_block.sorted_mi_blocks()
            if len(mi_blocks) <= 0:
                continue
            asm_block = mi_blocks[0].asm_block

            va = Addr(va=asm_block.va)
            yield Waypoint(start=va, end=va, asserts=[], replacements=[])


def trace_edges(cpg: CPG, session: Session, path: Path) -> Iterable[Edge]:
    return (session.query(cpg.Edge).get(edge_id) for edge_id in path.trace)


def trace_nodes(cpg: CPG, session: Session, cfg_path: Path) -> Iterable[Node]:
    trace = cfg_path.trace
    if len(trace) == 0:
        raise StopIteration
    # Yield the source of the first edge
    yield session.query(cpg.Edge).get(trace[0]).source_node
    # Yield the targets of each edge
    for edge in trace_edges(cpg, session, cfg_path):
        yield edge.target_node


def cfg_path_to_instructions(cpg: CPG, session: Session, cfg_path: Path) -> Iterable[Instruction]:
    """Lower an Instruction-level CFG path in the CPG to VA blocks."""
    for node in trace_nodes(cpg, session, cfg_path):
        if isinstance(node, cpg.Instruction):
            yield node



"""This module contains classes that represent tables in the Postgres DB, or derive from them."""

from __future__ import annotations

import io
import logging
import pathlib
import shutil
import sys
import tempfile
import uuid
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, TypeVar

import minio
from fastapi.encoders import jsonable_encoder
from sqlalchemy import Column, ForeignKey, ForeignKeyConstraint, Index, create_engine, orm
from sqlalchemy.dialects.postgresql import ARRAY, JSONB, array
from sqlalchemy.ext import compiler
from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy.ext.mutable import MutableDict, MutableList
from sqlalchemy.orm import mapper, relationship, sessionmaker
from sqlalchemy.schema import DDLElement, Table
from sqlalchemy.sql.elements import Grouping
from sqlalchemy.sql.expression import (
    FromClause,
    case,
    func,
    join,
    literal,
    null,
    select,
    text,
    true,
)
from sqlalchemy.types import Boolean, String

from mate.poi.poi_types import Analysis
from mate_common.assertions import mate_assert
from mate_common.error import MateError
from mate_common.models.analyses import (
    AnalysisTaskInfo,
    AnalysisTaskState,
    FlowFinderAnnotations,
    FlowFinderSnapshotInfo,
    GraphRequests,
    POIResultComplexity,
    POIResultInfo,
)
from mate_common.models.artifacts import ArtifactInformation, ArtifactKind
from mate_common.models.builds import BuildInformation, BuildOptions, BuildState
from mate_common.models.compilations import CompilationInformation, CompilationState, CompileOptions
from mate_common.models.cpg_types import (
    AST_EDGES,
    CONTROL_DEP_FORWARD,
    CONTROL_FLOW_FORWARD,
    DATA_FLOW_FORWARD,
    INFO_FLOW_FORWARD,
    LOCAL_CONTROL_FLOW_FORWARD,
    POINTS_TO,
    EdgeKind,
    NodeKind,
)
from mate_common.models.manticore import (
    MantiserveTaskInformation,
    MantiserveTaskKind,
    MantiserveTaskState,
)
from mate_query import storage
from mate_query.config import MATE_DEFAULT_EXPLORATION_BOUND
from mate_query.cpg.models.core.base import enum_column, uuid_column
from mate_query.cpg.models.core.cpg import BaseCPG
from mate_query.cpg.models.core.edge import BaseEdge
from mate_query.cpg.models.core.node import BaseNode

if TYPE_CHECKING:
    from typing import (
        IO,
        Any,
        AnyStr,
        Callable,
        ClassVar,
        Dict,
        Final,
        Iterator,
        List,
        Optional,
        Set,
        Tuple,
        Type,
    )

    from sqlalchemy.sql.expression import CTE, ColumnElement, Null

    from mate_query.cpg.models.core import Edge, Node


@as_declarative()
class Model:
    pass


logger = logging.getLogger(__name__)


# These are shared global instances
_connection_string: Optional[str] = None
_engine = None
_metadata = Model.metadata  # type: ignore[attr-defined]
_Session: Optional[Session] = None


class UninitializedDatabaseError(MateError):
    """An exception denoting an attempted use of an uninitialized database."""

    pass


# =================================================================


class Session(orm.Session):
    def graph_from_build(self, build: Build) -> Graph:
        return Graph.from_build(build, self)


# =================================================================


class Artifact:
    """Represents a filesystem artifact associated with zero or more builds.

    Artifacts are tagged with a ``kind``, which indicates their disposition within MATE. The
    ``kind`` may or may not reflect the file format itself. Artifacts also carry an ``attributes``
    dictionary, which contains unstructured metadata relevant to the artifact.
    """

    __table__: Table = None

    builds: List[Build]
    compilations: List[Compilation]

    def __init__(self, uuid: str, kind: ArtifactKind, *, attributes: Dict[str, Any] = {}) -> None:
        self.uuid = uuid
        self.kind = kind
        self.attributes = attributes

    @classmethod
    def create_with_object(
        cls, kind: ArtifactKind, *, fileobj: IO[AnyStr], attributes: Dict[str, Any] = {}
    ) -> Artifact:
        """Create a new ``Artifact`` and put the given fileobj into the object store, associating it
        with the newly created artifact."""
        artifact = cls(uuid=uuid.uuid4().hex, kind=kind, attributes=attributes)
        artifact.put_object(fileobj)
        return artifact

    def has_object(self) -> bool:
        """Returns whether this artifact has an associated object, i.e. whether an object has been
        upload to the object store for it."""
        try:
            storage.client.stat_object("artifacts", self.uuid)
        except minio.error.NoSuchKey:
            return False
        return True

    def put_object(self, fileobj: IO[AnyStr], **kwargs: Any) -> None:
        """Puts the given fileobj into the object store, associating it with this artifact."""

        # First annoying hack: `copyfileobj` doesn't handle mixed file modes well, so we need
        # to suss our our source's mode.
        if hasattr(fileobj, "mode"):
            if "b" in fileobj.mode:
                mode = "w+b"
            else:
                mode = "w+"
        elif isinstance(fileobj, io.StringIO):
            # Second annoying hack: the StringIO wrapper doesn't include a mode,
            # so we have to sniff its instance directly and specialize for it.
            mode = "w+"
        else:
            # If we don't have a mode attr, then we're probably dealing with a "file-like"
            # Response object or something similar. Treat these as binary-only for the time
            # being, which is almost certainly incorrect.
            mode = "w+b"

        with tempfile.TemporaryFile(mode=mode) as dst:
            shutil.copyfileobj(fileobj, dst)
            length = dst.tell()
            mate_assert(
                length != 0, f"artifact: given an empty fileobj to upload {self.uuid=} {self.kind=}"
            )

            dst.seek(0)

            # Third annoying hack: if we opened the `dst` tempfile in text mode above, we need
            # to grab its underlying binary buffer to put it into the object store.
            if hasattr(dst, "buffer"):
                real_dst = dst.buffer  # type: ignore[attr-defined]
            else:
                real_dst = dst

            storage.client.put_object("artifacts", self.uuid, real_dst, length, **kwargs)

    @contextmanager
    def get_object(self) -> Iterator[IO[bytes]]:
        """Yields the object associated with this artifact as a file-like object."""
        # NOTE(ww): For whatever reason, just yielding the response here doesn't work;
        # it fails silently and the response is empty. urllib3's HTTPResponse *claims*
        # to implement IO, so this is probably a bug on their side.
        try:
            response = storage.client.get_object("artifacts", self.uuid)
            dst = tempfile.TemporaryFile()
            shutil.copyfileobj(response, dst)
            dst.seek(0)
            yield dst
        finally:
            # Another probable (async) bug: this tempfile should be closed, but doing
            # so here causes a ChunkedEncodingError. Pain.
            # dst.close()
            response.close()
            # Needed? Which connection does this release? If it's the pooled one,
            # we probably don't want to release it here.
            # response.release_conn()

    def persist_locally(self, *, suffix: Optional[str] = None) -> pathlib.Path:
        """Create a local copy of the artifact and return it as a ``pathlib.Path``.

        The caller is responsible for disposing of the path. No cleanup is performed by default.
        """
        local_dir = pathlib.Path(tempfile.mkdtemp())
        local_file = local_dir / self.attributes["filename"]
        if suffix:
            local_file = local_file.with_suffix(suffix)

        with local_file.open("wb") as io:
            try:
                response = storage.client.get_object("artifacts", self.uuid)
                shutil.copyfileobj(response, io)
            finally:
                response.close()
        return local_file

    def to_info(self) -> ArtifactInformation:
        """Returns an ``ArtifactInformation`` model for this artifact."""
        return ArtifactInformation(
            artifact_id=self.uuid,
            kind=self.kind,
            has_object=self.has_object(),
            attributes=self.attributes,
            build_ids=[b.uuid for b in self.builds],
            compilation_ids=[c.uuid for c in self.compilations],
        )


class Compilation:
    __table__: Table = None

    # NOTE(ww): Give mypy a helping hand with these relationship properties.
    artifacts: List[Artifact]
    builds: List[Build]

    def __init__(
        self,
        uuid: str,
        state: CompilationState,
        source_artifact: Artifact,
        *,
        options: Dict[str, Any],  # TODO(ww): This should become CompileOptions.
    ) -> None:
        mate_assert(
            source_artifact.kind.is_compile_target(),
            f"API misuse: {source_artifact.kind=} is not a compilation target",
        )

        self.uuid = uuid
        self.state = state
        self.source_artifact = source_artifact
        self.options = options

    @classmethod
    def create(cls, source_artifact: Artifact, *, options: Dict[str, Any]) -> Compilation:
        return cls(
            uuid=uuid.uuid4().hex,
            state=CompilationState.Created,
            source_artifact=source_artifact,
            options=options,
        )

    def transition_to_state(self, new_state: CompilationState) -> None:
        mate_assert(
            self.state.can_transition_to(new_state),
            f"impossible compilation state transition: {self.state} -> {new_state}",
        )
        self.state = new_state

    @cached_property
    def challenge_id(self) -> Optional[str]:
        """Returns the challenge broker ID that corresponds to this compilation, if this compilation
        was created from a brokered challenge."""
        return self.source_artifact.attributes.get("challenge_id")

    def to_info(self) -> CompilationInformation:
        """Returns a ``CompilationInformation`` model for this compilation."""
        log_artifact = next(
            (a for a in self.artifacts if a.kind == ArtifactKind.CompileOutputCompileLog), None
        )

        return CompilationInformation(
            compilation_id=self.uuid,
            build_ids=[b.uuid for b in self.builds],
            state=self.state,
            source_artifact=self.source_artifact.to_info(),
            log_artifact=log_artifact.to_info() if log_artifact else None,
            artifact_ids=[a.uuid for a in self.artifacts],
            options=CompileOptions(**self.options),
        )


class Build:
    __table__: Table = None

    # NOTE(ww): Same as Compilation; help mypy out.
    artifacts: List[Artifact]
    analysis_tasks: List[AnalysisTask]
    mantiserve_tasks: List[MantiserveTask]

    def __init__(
        self,
        uuid: str,
        state: BuildState,
        bitcode_artifact: Artifact,
        compilation: Compilation,
        *,
        options: Dict[str, Any],  # TODO(ww): This should become BuildOptions.
        attributes: Dict[str, Any],
    ) -> None:
        mate_assert(
            bitcode_artifact.kind == ArtifactKind.CompileOutputBitcode,
            f"API misuse: {bitcode_artifact.kind=} != CompileOutputBitcode",
        )

        self.uuid = uuid
        self.state = state
        self.bitcode_artifact = bitcode_artifact
        self.compilation = compilation
        self.options = options
        self.attributes = attributes

    @classmethod
    def create(
        cls,
        bitcode_artifact: Artifact,
        compilation: Compilation,
        *,
        options: Dict[str, Any],
        attributes: Dict[str, Any] = {},
    ) -> Build:
        return cls(
            uuid=uuid.uuid4().hex,
            state=BuildState.Created,
            bitcode_artifact=bitcode_artifact,
            compilation=compilation,
            options=options,
            attributes=attributes,
        )

    def transition_to_state(self, new_state: BuildState) -> None:
        mate_assert(
            self.state.can_transition_to(new_state),
            f"impossible build state transition: {self.state} -> {new_state}",
        )
        self.state = new_state

    def to_info(self, artifact_detail: bool = False) -> BuildInformation:
        """Returns a ``BuildInformation`` model for this build.

        If ``artifact_detail`` is supplied, the resulting ``BuildInformation`` includes the
        ``ArtifactInformation`` for each artifact in the build.
        """
        return BuildInformation(
            build_id=self.uuid,
            compilation=self.compilation.to_info(),
            state=self.state,
            bitcode_artifact=self.bitcode_artifact.to_info(),
            artifact_ids=[a.uuid for a in self.artifacts],
            artifacts=([a.to_info() for a in self.artifacts] if artifact_detail else []),
            analysis_task_ids=[a.uuid for a in self.analysis_tasks],
            mantiserve_task_ids=[t.uuid for t in self.mantiserve_tasks],
            options=BuildOptions(**self.options),
            attributes=self.attributes,
        )


class CompilationArtifactAssociation:
    """A many-many association table for mapping artifacts to compilations and compilations to
    artifacts.

    An artifact can have many compilations, since multiple independent ``db.Compilation`s
    can be created (with different configurations) for one artifact.

    Each ``db.Compilation``, in turn, can have many ``db.Artifacts`` (in the form
    of compilation products and byproducts).
    """

    __table__: Table = None


class BuildArtifactAssociation:
    """A many-many association table for mapping artifacts to builds and builds to artifacts.

    An artifact can have many builds in one case: a compile target that's
    used to produce multiple builds.

    A build can have many artifacts in the normal case: each build is expected
    to produce compilation outputs and other artifacts that are associated with it.
    """

    __table__: Table = None


class MantiserveTask:
    """Mantiserve task info."""

    __table__: Table = None

    # NOTE(ekilmer): We declare a variable attribute here to help mypy.
    build: Build
    artifacts: List[Artifact]

    def __init__(
        self,
        uuid: str,
        build: Build,
        kind: MantiserveTaskKind,
        request_msg: Dict[str, Any],
        state: MantiserveTaskState,
        response_msg: Optional[Dict[str, Any]] = None,
        docker_image: Optional[str] = None,
        job_id: Optional[str] = None,
    ):
        """request_msg and response_msg are a json-encoded response message related to
        MantiserveTaskKind."""
        self.uuid = uuid
        self.build = build
        self.kind = kind
        self.request_msg = request_msg
        self.response_msg = response_msg
        self.state = state
        self.docker_image = docker_image
        self.job_id = job_id

    @classmethod
    def create(
        cls,
        build: Build,
        kind: MantiserveTaskKind,
        request_msg: Dict[str, Any],
        docker_image: Optional[str] = None,
    ) -> MantiserveTask:
        return MantiserveTask(
            uuid=uuid.uuid4().hex,
            build=build,
            kind=kind,
            request_msg=request_msg,
            response_msg=None,
            state=MantiserveTaskState.Created,
            docker_image=docker_image,
            job_id=None,
        )

    def transition_to_state(
        self, new_state: MantiserveTaskState, msg: Optional[str] = None
    ) -> None:
        """Transition to a different state with validity checking.

        :param new_state: Transition to this state
        :param msg: Optional log message to attach to ``self.response_msg``. If
            not present, does not touch ``self.response_msg``. Should only be used
            when transitioning to a `MantiserveTaskState.Failed` state.
        :return: None
        """
        mate_assert(
            self.state.can_transition_to(new_state),
            f"impossible mantiserve task state transition: {self.state} -> {new_state}",
        )
        if msg is not None:
            if self.response_msg is not None:
                logger.warning(
                    "Overwriting previous response message while transitioning state from "
                    f"'{self.state}' to '{new_state}':\n"
                    f"\tOriginal: {self.response_msg}\n"
                    f"\tNew: {msg}"
                )
            self.response_msg = {"log": msg}
        self.state = new_state

    def to_info(self) -> MantiserveTaskInformation:
        """Returns a ``MantiserveTaskInformation`` model for this Mantiserve task."""
        return MantiserveTaskInformation(
            task_id=self.uuid,
            build_id=self.build.uuid,
            artifact_ids=[a.uuid for a in self.artifacts],
            kind=self.kind,
            request=self.request_msg,
            result=self.response_msg,
            state=self.state,
            docker_image=self.docker_image,
        )


class MantiserveTaskArtifactAssociation:
    """Associate table of a MantiserveTask to many Artifacts."""

    __table__: Table = None


# =================================================================


class AnalysisTask:
    """Represents a concrete invocation of a POI analysis on a ``Build``."""

    __table__: Table = None

    build: Build
    poi_results: List[POIResult]

    def __init__(self, uuid: str, analysis_name: str, state: AnalysisTaskState, build: Build):
        self.uuid = uuid
        self.analysis_name = analysis_name
        self.state = state
        self.build = build

    @classmethod
    def create(cls, analysis_name: str, build: Build) -> AnalysisTask:
        return cls(
            uuid=uuid.uuid4().hex,
            analysis_name=analysis_name,
            state=AnalysisTaskState.Created,
            build=build,
        )

    def transition_to_state(self, new_state: AnalysisTaskState) -> None:
        mate_assert(
            self.state.can_transition_to(new_state),
            f"impossible analysis task state transition: {self.state} -> {new_state}",
        )
        self.state = new_state

    def to_info(self) -> AnalysisTaskInfo:
        """Returns an ``AnalysisTaskInfo`` model for this analysis task."""
        return AnalysisTaskInfo(
            analysis_task_id=self.uuid,
            analysis_name=self.analysis_name,
            build_id=self.build.uuid,
            poi_result_ids=[poi.uuid for poi in self.poi_results],
            state=self.state,
        )


class POIResult:
    """Represents a single POI result, produced by running a POI analysis through an
    ``AnalysisTask``."""

    __table__: Table = None

    analysis_task: AnalysisTask
    parent: Optional[POIResult]
    children: List[POIResult]
    snapshots: List[FlowFinderSnapshot]

    def __init__(
        self,
        uuid: str,
        poi: Dict[str, Any],
        analysis_task: AnalysisTask,
        graph_requests: GraphRequests,
        flagged: bool = False,
        done: bool = False,
        complexity: POIResultComplexity = POIResultComplexity.Unknown,
    ) -> None:
        self.uuid = uuid
        self.poi = poi
        self.analysis_task = analysis_task
        self.graph_requests = graph_requests
        self.flagged = flagged
        self.done = done
        self.complexity = complexity

    @classmethod
    def create(
        cls,
        poi: Dict[str, Any],
        analysis_task: AnalysisTask,
        graph_requests: List[Dict[str, Any]],
        complexity: POIResultComplexity = POIResultComplexity.Unknown,
    ) -> POIResult:
        return cls(
            uuid=uuid.uuid4().hex,
            poi=poi,
            analysis_task=analysis_task,
            graph_requests=graph_requests,
            complexity=complexity,
        )

    def to_info(self) -> POIResultInfo:
        """Returns a ``POIResultInfo`` model for this POI result."""
        return POIResultInfo(
            poi_result_id=self.uuid,
            build_id=self.analysis_task.build.uuid,
            analysis_task_id=self.analysis_task.uuid,
            analysis_name=self.analysis_task.analysis_name,
            poi=self.poi,
            flagged=self.flagged,
            done=self.done,
            complexity=self.complexity,
            parent_result_id=self.parent.uuid if self.parent else None,
            child_result_ids=[c.uuid for c in self.children],
            snapshot_ids=[s.uuid for s in self.snapshots],
            graph_requests=self.graph_requests,
            insight=self.poi["insight"],
            background=Analysis.by_name(self.analysis_task.analysis_name).background(),
            salient_functions=self.poi.pop("salient_functions"),
        )


class FlowFinderSnapshot:
    """Represents the state of a user's analysis of a ``POIResult``."""

    __table__: Table = None

    poi_result: Optional[POIResult]
    build: Build

    def __init__(
        self,
        *,
        uuid: str,
        poi_result: Optional[POIResult],
        build: Build,
        label: str,
        filters: List[str],
        graph_requests: GraphRequests,
        hidden_graph_ids: List[str],
        hidden_node_ids: List[str],
        user_annotations: Dict[str, FlowFinderAnnotations],
    ) -> None:
        if poi_result is not None:
            mate_assert(
                poi_result.analysis_task.build.uuid == build.uuid,
                "invariant: POI result and build must match",
            )

        self.uuid = uuid
        self.poi_result = poi_result
        self.build = build
        self.label = label
        self.filters = filters
        self.graph_requests = jsonable_encoder(graph_requests)
        self.hidden_graph_ids = hidden_graph_ids
        self.hidden_node_ids = hidden_node_ids
        self.user_annotations = user_annotations

    @classmethod
    def create(
        cls,
        *,
        poi_result: Optional[POIResult],
        build: Build,
        label: str,
        filters: List[str] = [],
        graph_requests: GraphRequests = [],
        hidden_graph_ids: List[str] = [],
        hidden_node_ids: List[str] = [],
        user_annotations: Dict[str, FlowFinderAnnotations] = {},
    ) -> FlowFinderSnapshot:
        return cls(
            uuid=uuid.uuid4().hex,
            poi_result=poi_result,
            build=build,
            label=label,
            filters=filters,
            graph_requests=graph_requests,
            hidden_graph_ids=hidden_graph_ids,
            hidden_node_ids=hidden_node_ids,
            user_annotations=user_annotations,
        )

    def to_info(self) -> FlowFinderSnapshotInfo:
        """Returns a ``FlowFinderSnapshotInfo`` model for this snapshot."""
        return FlowFinderSnapshotInfo(
            snapshot_id=self.uuid,
            poi_result_id=self.poi_result.uuid if self.poi_result else None,
            build_id=self.build.uuid,
            label=self.label,
            filters=self.filters,
            graph_requests=self.graph_requests,
            hidden_graph_ids=self.hidden_graph_ids,
            hidden_node_ids=self.hidden_node_ids,
            user_annotations=self.user_annotations,
        )


# =================================================================


@lru_cache(maxsize=None)
def build_model_classes(build: str) -> Tuple[Type[BaseNode], Type[BaseEdge]]:
    class Node(Model, BaseNode):

        __tablename__ = f"base_nodes_{build}"
        build_id = build

        def __init__(self, uuid: str, kind: str, attributes: Dict[Any, Any]) -> None:
            self.uuid = uuid
            self.kind = kind
            self.attributes = attributes

    Node.__table__.create(_engine, checkfirst=True)
    NodeClass = Node

    class Edge(Model, BaseEdge):

        __tablename__ = f"base_edges_{build}"

        Node = NodeClass  # type: ignore[assignment]
        build_id = build

        def __init__(
            self,
            uuid: str,
            kind: str,
            source: str,
            target: str,
            attributes: Dict[Any, Any],
        ) -> None:
            super().__init__()  # type: ignore[call-arg]

            self.uuid = uuid
            self.kind = kind
            self.source = source
            self.target = target
            self.attributes = attributes

    Edge.__table__.create(_engine, checkfirst=True)
    EdgeClass = Edge

    return NodeClass, EdgeClass


# =================================================================


class CreateView(DDLElement):
    def __init__(self, name: str, selectable: FromClause) -> None:
        self.name = name
        self.selectable = selectable


@compiler.compiles(CreateView)
def emit_view(element: Any, compiler: Any, **_kw: Dict[str, Any]) -> str:
    return "CREATE OR REPLACE VIEW %s AS %s" % (
        element.name,
        compiler.sql_compiler.process(element.selectable, literal_binds=True),
    )


# =================================================================

# AdvisoryLockBuildTransaction takes a Postgres advisory lock for the
# current transaction until it is commited or aborted.
# AdvisoryLockBuildTransaction may use the same lock id for two
# different UUIDs, since the space of possible UUIDs is larger than
# the space of possible lock IDs. As a result, it should not be used
# to guard operations that require synchronized access to more than
# one UUID at a time.
class AdvisoryLockBuildTransaction(DDLElement):
    def __init__(self, build: str) -> None:
        self.build = build


@compiler.compiles(AdvisoryLockBuildTransaction)
def emit_lock(element: Any, _compiler: Any, **_kw: Dict[str, Any]) -> str:
    # convert build uuid to BIGINT key
    # Note: only uses partial uuid.
    lockid = int.from_bytes(
        int(element.build[0:16], 16).to_bytes(length=8, byteorder=sys.byteorder),
        byteorder=sys.byteorder,
        signed=True,
    )
    return f"SELECT pg_advisory_xact_lock({lockid})"


# =================================================================


def new_session() -> orm.Session:
    if _Session is None:
        raise UninitializedDatabaseError()
    return _Session()


@contextmanager
def session_scope() -> Iterator[orm.Session]:
    if _Session is None:
        raise UninitializedDatabaseError()
    session = _Session()
    try:
        yield session
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def initialize(connection_string: str, echo: bool = False, create: bool = False) -> None:
    global _connection_string, _engine, _metadata, _Session

    if _connection_string is not None:
        logger.error(
            f"MATE has already been initialized with a connection string {_connection_string}. "
            "Skipping reinitialization."
        )
        return

    _connection_string = connection_string
    _engine = create_engine(
        connection_string,
        echo=echo,
        pool_pre_ping=True,
        connect_args={
            "keepalives": 1,
            "keepalives_idle": 30,
            "keepalives_interval": 10,
            "keepalives_count": 5,
        },
        executemany_mode="batch",
        # NOTE(ww): Don't place any limits on connections at the SQLAlchemy level.
        # Why not? Because SQLAlchemy's pooling strategies are a *little* too
        # clever for our use case: by default it will accept N connections in
        # the pool, complemented by M "overflow" connections when those N
        # connections are all active. When that M threshold is crossed, it
        # begins rejecting connections. This is normally desirable, but MATE's
        # access patterns can include dozens of long-running connections when
        # the system is under heavy load (e.g., multiple simultaneous CPG builds).
        # We don't want in-progress builds to cause knock-on failures, so we
        # tell SQLAlchemy to create as many "overflow" connections as the
        # underlying database will tolerate.
        # See: https://gitlab-ext.galois.com/mate/MATE/-/merge_requests/1573
        max_overflow=-1,
    )
    _Session = sessionmaker(bind=_engine, class_=Session)

    Artifact.__table__ = _make_artifact_table("artifacts")
    mapper(
        Artifact,
        Artifact.__table__,
        properties={
            "compilations": relationship(
                Compilation,
                secondary=lambda: CompilationArtifactAssociation.__table__,
                back_populates="artifacts",
            ),
            "builds": relationship(
                Build,
                secondary=lambda: BuildArtifactAssociation.__table__,
                back_populates="artifacts",
            ),
        },
    )

    Compilation.__table__ = _make_compilation_table("compilations")
    mapper(
        Compilation,
        Compilation.__table__,
        properties={
            "source_artifact": relationship(Artifact),
            "artifacts": relationship(
                Artifact,
                secondary=lambda: CompilationArtifactAssociation.__table__,
                back_populates="compilations",
            ),
            "builds": relationship(
                Build,
                backref="compilation",
            ),
        },
    )

    Build.__table__ = _make_build_table("builds")
    mapper(
        Build,
        Build.__table__,
        properties={
            "bitcode_artifact": relationship(Artifact),
            "artifacts": relationship(
                Artifact,
                secondary=lambda: BuildArtifactAssociation.__table__,
                back_populates="builds",
            ),
            "mantiserve_tasks": relationship(MantiserveTask, backref="build"),
            "snapshots": relationship(FlowFinderSnapshot, backref="build"),
        },
    )

    CompilationArtifactAssociation.__table__ = _make_compilation_artifact_association_table(
        "compilation_artifact_association"
    )
    mapper(CompilationArtifactAssociation, CompilationArtifactAssociation.__table__)

    BuildArtifactAssociation.__table__ = _make_build_artifact_association_table(
        "build_artifact_association"
    )
    mapper(BuildArtifactAssociation, BuildArtifactAssociation.__table__)

    AnalysisTask.__table__ = _make_analysis_task_table("analysis_tasks")
    mapper(
        AnalysisTask,
        AnalysisTask.__table__,
        properties={
            "build": relationship(Build, backref="analysis_tasks"),
        },
    )

    POIResult.__table__ = _make_poi_results_table("poi_results")
    mapper(
        POIResult,
        POIResult.__table__,
        properties={
            "analysis_task": relationship(
                AnalysisTask,
                backref="poi_results",
            ),
            "parent": relationship(
                POIResult, backref="children", remote_side=lambda: POIResult.uuid
            ),
            "snapshots": relationship(FlowFinderSnapshot, backref="poi_result"),
        },
    )

    FlowFinderSnapshot.__table__ = _make_flowfinder_snapshots_table("flowfinder_snapshots")
    mapper(
        FlowFinderSnapshot,
        FlowFinderSnapshot.__table__,
    )

    MantiserveTask.__table__ = _make_mantiserve_table("mantiserve_task")
    mapper(
        MantiserveTask,
        MantiserveTask.__table__,
        properties={
            "artifacts": relationship(
                Artifact,
                secondary=lambda: MantiserveTaskArtifactAssociation.__table__,
                backref="mantiserve_task",
            ),
        },
    )
    MantiserveTaskArtifactAssociation.__table__ = _make_mantiserve_task_artifact_association_table(
        "mantiserve_task_artifact_association"
    )
    mapper(MantiserveTaskArtifactAssociation, MantiserveTaskArtifactAssociation.__table__)

    if not create:
        return

    # NOTE(ww): SQLAlchemy's `create_all` is idempotent by default,
    # and shouldn't fail under normal circumstances (even if the tables
    # already exist). However, it *can* fail on a race condition
    # where two services are attempting to create the same tables at
    # exactly the same time. At least one of those services should succeed
    # and not raise during creation, so we turn it into a warning here.
    # See: https://gitlab-ext.galois.com/mate/MATE/-/issues/1181
    try:
        _metadata.create_all(bind=_engine)
    except Exception as e:
        logger.warning(f"table creation failed: {e}")


def _make_artifact_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        uuid_column(),
        enum_column("kind", ArtifactKind),
        Column("attributes", MutableDict.as_mutable(JSONB), nullable=False),
    )


def _make_compilation_artifact_association_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        Column(
            "compilation_uuid",
            String,
            ForeignKey("compilations.uuid"),
            primary_key=True,
            nullable=False,
        ),
        Column(
            "artifact_uuid", String, ForeignKey("artifacts.uuid"), primary_key=True, nullable=False
        ),
    )


def _make_build_artifact_association_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        Column("build_uuid", String, ForeignKey("builds.uuid"), primary_key=True, nullable=False),
        Column(
            "artifact_uuid", String, ForeignKey("artifacts.uuid"), primary_key=True, nullable=False
        ),
    )


def _make_compilation_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        uuid_column(),
        enum_column("state", CompilationState),
        Column("source_artifact_uuid", String, ForeignKey("artifacts.uuid"), nullable=False),
        Column("options", JSONB, nullable=False),
    )


def _make_build_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        uuid_column(),
        enum_column("state", BuildState),
        Column("bitcode_artifact_uuid", String, ForeignKey("artifacts.uuid"), nullable=False),
        Column("compilation_uuid", String, ForeignKey("compilations.uuid"), nullable=False),
        Column("options", JSONB, nullable=False),
        Column("attributes", MutableDict.as_mutable(JSONB), nullable=False),
    )


def _make_analysis_task_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        uuid_column(),
        Column("analysis_name", String, nullable=False),
        enum_column("state", AnalysisTaskState),
        Column("build_uuid", String, ForeignKey("builds.uuid"), nullable=False),
    )


def _make_poi_results_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        uuid_column(),
        Column("poi", JSONB, nullable=False),
        Column("analysis_task_uuid", String, ForeignKey("analysis_tasks.uuid"), nullable=False),
        Column("graph_requests", JSONB, nullable=False),
        Column("flagged", Boolean, nullable=False),
        Column("done", Boolean, nullable=False),
        Column("parent_uuid", String, ForeignKey("poi_results.uuid"), nullable=True),
        enum_column("complexity", POIResultComplexity),
    )


def _make_flowfinder_snapshots_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        uuid_column(),
        Column("poi_result_uuid", String, ForeignKey("poi_results.uuid"), nullable=True),
        Column("build_id", String, ForeignKey("builds.uuid"), nullable=True),
        Column("label", String, nullable=False, unique=False),
        Column("filters", MutableList.as_mutable(JSONB), nullable=False),
        Column("graph_requests", MutableList.as_mutable(JSONB), nullable=False),
        Column("hidden_graph_ids", MutableList.as_mutable(JSONB), nullable=False),
        Column("hidden_node_ids", MutableList.as_mutable(JSONB), nullable=False),
        Column("user_annotations", MutableDict.as_mutable(JSONB), nullable=False),
    )


def _make_mantiserve_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        uuid_column(),
        Column("build_uuid", String, ForeignKey("builds.uuid"), nullable=False),
        enum_column("kind", MantiserveTaskKind),
        Column("request_msg", JSONB, nullable=False),
        Column("response_msg", MutableDict.as_mutable(JSONB), nullable=True),
        enum_column("state", MantiserveTaskState),
        Column("docker_image", String, nullable=True),
        Column("job_id", String, nullable=True),
    )


def _make_mantiserve_task_artifact_association_table(name: str) -> Table:
    return Table(
        name,
        _metadata,
        Column(
            "mantiserve_task_uuid",
            String,
            ForeignKey("mantiserve_task.uuid"),
            primary_key=True,
            nullable=False,
        ),
        Column(
            "artifact_uuid",
            String,
            ForeignKey("artifacts.uuid"),
            primary_key=True,
            nullable=False,
            unique=True,
        ),
    )


# =================================================================


class Graph(BaseCPG):
    def __init__(
        self,
        parent: Optional[Graph],
        name: str,
        build: str,
        selectables: Optional[Tuple[FromClause, FromClause]],
        session: Session,
    ) -> None:
        """Initialize a graph, setting internal state for use as a context manager.

        This constructor should generally not be called directly; instead, users should call one of
        the named classmethod constructors.
        """
        self.name = name
        self.parent = parent
        self.build = build
        self.session = session

        session.execute(AdvisoryLockBuildTransaction(build))
        if parent is None:
            self.BaseNode, self.BaseEdge = build_model_classes(build)
            session.commit()
            self.node_selectable = self.BaseNode.__table__.select()
            self.edge_selectable = self.BaseEdge.__table__.select()
        else:
            self.BaseNode = self.parent.BaseNode  # type: ignore[union-attr]
            self.BaseEdge = self.parent.BaseEdge  # type: ignore[union-attr]
            self.node_selectable, self.edge_selectable = selectables  # type: ignore[misc]

        node_table_name = f"nodes_{name}"
        edge_table_name = f"edges_{name}"

        self.session.bind.execute(CreateView(node_table_name, self.node_selectable))
        self.session.bind.execute(CreateView(edge_table_name, self.edge_selectable))
        session.commit()

        # NOTE(lb) These don't inherit from Node, Edge directly because you
        # can't inherit from something with a different __tablename__
        self.Node = type(
            f"Node{self.name}",
            (Model, BaseNode),
            {"cpg": self, "build_id": build, "__tablename__": node_table_name},
        )
        self.Edge = type(
            f"Edge{self.name}",
            (Model, BaseEdge),
            {"cpg": self, "build_id": build, "Node": self.Node, "__tablename__": edge_table_name},
        )

        self._attach_node_models(self.name)

    @classmethod
    def from_build(cls, build: Build, session: Session) -> Graph:
        """Create a graph consisting of all nodes/edges from an existing build."""
        return cls(
            None,
            uuid.uuid4().hex,
            build.uuid,
            None,
            session,
        )

    def subgraph(self, builder: SubgraphBuilder) -> Graph:
        """Run the given SubgraphBuilder on the current graph."""
        return builder.build(self)

    def paths(self, builder: PathBuilder) -> Type[Path]:
        """Run the given PathBuilder on the current graph."""
        return builder.build(self)

    @cached_property
    def cfg(self) -> Graph:
        """Construct the control-flow subgraph of the current graph."""
        return self.subgraph(SubgraphBuilder().with_edge_kinds(*LOCAL_CONTROL_FLOW_FORWARD))

    @cached_property
    def icfg(self) -> Graph:
        """Construct the interprocedural control-flow subgraph of the current graph."""
        return self.subgraph(SubgraphBuilder().with_edge_kinds(*CONTROL_FLOW_FORWARD))

    @cached_property
    def cg(self) -> Graph:
        """Construct the call-subgraph of the current graph."""
        return self.subgraph(
            SubgraphBuilder().with_node_kinds(NodeKind.FUNCTION).with_edge_kinds(EdgeKind.CALLGRAPH)
        )

    @cached_property
    def ast(self) -> Graph:
        """Construct the abstract sytnax tree subgraph of the current graph."""
        return self.subgraph(SubgraphBuilder().with_edge_kinds(*AST_EDGES))

    @cached_property
    def dfg(self) -> Graph:
        """Construct the data-flow subgraph of the current graph."""
        return self.subgraph(SubgraphBuilder().with_edge_kinds(*DATA_FLOW_FORWARD))

    @cached_property
    def ifg(self) -> Graph:
        """Construct the information-flow subgraph of the current graph."""
        return self.subgraph(SubgraphBuilder().with_edge_kinds(*INFO_FLOW_FORWARD))

    @cached_property
    def cdg(self) -> Graph:
        """Construct the control-dependence subgraph of the current graph."""
        return self.subgraph(SubgraphBuilder().with_edge_kinds(*CONTROL_DEP_FORWARD))

    @cached_property
    def ptg(self) -> Graph:
        """Construct the points-to subgraph of the current graph."""
        return self.subgraph(SubgraphBuilder().with_edge_kinds(*POINTS_TO))


class SubgraphBuilder:
    """A builder class for subgraphs."""

    def __init__(self) -> None:
        self.node_kinds: Set[NodeKind] = set()
        self.edge_kinds: Set[EdgeKind] = set()

    def with_node_kinds(self, *kinds: NodeKind) -> SubgraphBuilder:
        """Specify a set of node kinds to be included within the subgraph."""
        self.node_kinds |= set(kinds)
        return self

    def with_edge_kinds(self, *kinds: EdgeKind) -> SubgraphBuilder:
        """Specify a set of edge kinds to be included within the subgraph."""
        self.edge_kinds |= set(kinds)
        return self

    def build(self, graph: Graph) -> Graph:
        """Concretize the subgraph defined by this builder, with respect to the given graph."""
        self.session = graph.session

        node_alias = graph.node_selectable.alias()
        edge_alias = graph.edge_selectable.alias()

        subgraph_node_selectable = node_alias.select(
            (node_alias.c.kind.in_([kind.value for kind in self.node_kinds]))
            if self.node_kinds != set()
            else true()
        ).alias()

        subgraph_edge_selectable = edge_alias.select(
            (
                (edge_alias.c.kind.in_([kind.value for kind in self.edge_kinds]))
                if self.edge_kinds != set()
                else true()
            )
            & edge_alias.c.source.in_(select([subgraph_node_selectable.c.uuid]))
            & edge_alias.c.target.in_(select([subgraph_node_selectable.c.uuid]))
        )

        return Graph(
            graph,
            uuid.uuid4().hex,
            graph.build,
            (subgraph_node_selectable, subgraph_edge_selectable),
            graph.session,
        )


@dataclass
class Path:
    """For documentation of this class's attributes, see comments in ``_make_path_table`` or use
    ``help`` at runtime."""

    __table__ = None

    info: Any
    source: Any
    source_stack: Tuple[Any, ...]
    target: Any
    edge: Any
    trace: Tuple[Any, ...]
    state: Any
    stack: Tuple[Any, ...]
    stack_top: Any
    c: Any
    edge_symbol_table: Optional[Table] = None
    transition_table: Optional[Table] = None
    is_reversed: ClassVar[bool] = False
    stepped_over_functions: ClassVar[set[str]] = field(default_factory=set)
    cfl: ClassVar[bool] = False

    @classmethod
    def populate_transition_table(cls, graph: Graph) -> None:
        raise NotImplementedError

    @classmethod
    def populate_edge_symbol_table(cls, graph: Graph) -> None:
        raise NotImplementedError

    @classmethod
    def initial_state(cls) -> Null:
        return null()

    @classmethod
    def initial_stack(cls) -> Null:
        return null()


def _make_path_table(name: str, node_table: Table, edge_table: Table) -> Table:
    info_comment = "User-defined distinguisher for this path if provided (an extra piece of data that can then be consulted e.g. in a continuing_while predicate)"
    source_comment = "The source node of this path (if reversed or keep_start=True)"
    source_stack_comment = "The stack this path was initialized in (if keep_start=True)"
    target_comment = "The target node of this path (if not reversed or keep_start=True)"
    trace_comment = "The trace of edges visited by this path (if keep_trace=True)"
    state_comment = "The current PDA state of this path (if CFL query)"
    stack_comment = "The current PDA stack of this path (if CFL query)"
    stack_top_comment = "The top of the current PDA stack of this path (if CFL query)"
    edge_comment = "The last visited edge of this path (if keep_edge=True)"
    return Table(
        name,
        _metadata,
        Column(
            "info",
            String,
            primary_key=True,
            nullable=True,
            index=True,
            doc=info_comment,
            comment=info_comment,
        ),
        Column("source", String, primary_key=True, doc=source_comment, comment=source_comment),
        Column(
            "source_stack",
            ARRAY(String, as_tuple=True),
            primary_key=True,
            doc=source_stack_comment,
            comment=source_stack_comment,
        ),
        Column(
            "target",
            String,
            primary_key=True,
            index=True,
            doc=target_comment,
            comment=target_comment,
        ),
        Column(
            "trace",
            ARRAY(String, as_tuple=True),
            primary_key=True,
            doc=trace_comment,
            comment=trace_comment,
        ),
        Column(
            "state", String, primary_key=True, index=True, doc=state_comment, comment=state_comment
        ),
        Column(
            "stack",
            ARRAY(String, as_tuple=True),
            primary_key=True,
            index=True,
            doc=stack_comment,
            comment=stack_comment,
        ),
        Column(
            "stack_top",
            String,
            primary_key=True,
            index=True,
            doc=stack_top_comment,
            comment=stack_top_comment,
        ),
        Column(
            "edge",
            String,
            primary_key=True,
            nullable=True,
            index=True,
            doc=edge_comment,
            comment=edge_comment,
        ),
        ForeignKeyConstraint(["source"], [node_table.c.uuid]),
        ForeignKeyConstraint(["target"], [node_table.c.uuid]),
        ForeignKeyConstraint(["edge"], [edge_table.c.uuid]),
    )


# HACK: Force postgres to run joins in order written
#
# This was previously necessary for ensuring that the PDA transitions
# join on the transition table, *then* the edge table, see also PathBuilder.
@contextmanager
def join_collapse_limit(session: orm.Session) -> Iterator[None]:
    session.execute("SET join_collapse_limit=1")
    try:
        yield
    finally:
        session.execute("RESET join_collapse_limit")


def set_n_distinct(
    session: orm.Session, tablename: str, columns: Set[str], n_distinct: int
) -> None:
    to_update = set()
    for column in columns:
        needs_update = (
            f"SELECT 1 FROM pg_stats WHERE tablename = '{tablename}' AND"
            f" attname = '{column}' AND n_distinct = {n_distinct}::int"
        )
        if session.execute(text(needs_update)).first() is None:
            logger.debug(f"{tablename}.{column} n_distinct requires update")
            to_update.add(column)
        else:
            logger.debug(f"{tablename}.{column} n_distinct does not require update")

    for column in to_update:
        logger.debug(f"{tablename}.{column} (n_distinct={n_distinct})")
        alter_cmd = f"ALTER TABLE {tablename} ALTER COLUMN {column} SET (n_distinct={n_distinct})"
        logger.debug(alter_cmd)
        session.execute(text(alter_cmd))

    if len(to_update) > 0:
        # session.commit()
        analyze_cmd = f"ANALYZE {tablename}"
        logger.debug(analyze_cmd)
        session.execute(text(analyze_cmd))
        logger.debug(f"n_distinct statistics for {tablename} ({', '.join(to_update)}) updated")
    else:
        logger.debug(f"n_distinct statistics for {tablename} did not require update")


PB = TypeVar("PB", bound="PathBuilder")

# Bottom of the PDA stack, see PathBuilder

BOT: Final[str] = "$"


class PathBuilder:
    """A builder class for enumerating paths in a graph.

    PathBuilder supports "CFL-reachability queries" (CFL stands for
    "context-free language"). These queries have two parts:

    1. In a pre-processing step, each edge in the graph is labeled with some
       symbol (string).
    2. The graph is explored one path at a time by a push-down automaton (PDA)
       that reads the edge labels and recognizes strings in a context-free
       grammar.

    These queries return the set of all paths in the graph (that start at some
    set of nodes) where the edge labels on that path form a string in the
    context-free language recognized by the PDA.

    For example the "balanced parentheses" context-free language can be used to
    encode the problem of finding control-flow graph (CFG) paths that have a
    balanced number of calls and returns.

    Step (2) above can be accomplished by any number of specialized algorithms,
    but the PDA approach is relatively simple and it's the only one MATE has so
    far.

    The state of the PDA and its set of transitions are kept in a dedicated
    database tables, see ``_make_path_table`` and ``_make_transition_table``
    respectively, and the comments on their columns especially.

    To perform a transition, the PDA state table is first joined against its
    transition table, and then joined against the edge symbol table. This join
    order is more performant than vice-versa, because the edge table is
    typically way bigger. At each step, the automation can make multiple
    transitions (end up in multiple new states).

    Because the PDA state is kept in a table, and a table is like a set of
    tuples (configurations), we get some cycle-avoidance for free. However,
    this can get more complicated if the stack or trace is kept as part of the
    PDA configuration/state, because while these are theoretically bounded in
    practice they can be too big. See the ``exploration_bound`` parameter to
    ``build``.
    """

    def __init__(self, PathBase: Type[Path] = Path):
        self.PathBase = PathBase
        self.start_predicates: List[Callable[[Type[Node]], ColumnElement]] = []
        self.stop_predicates: List[Callable[[Type[Node]], ColumnElement]] = []
        self.continue_predicates: List[Callable[[Type[Path], Type[Edge]], ColumnElement]] = []
        self.no_detail_continue_predicates: List[
            Callable[[Type[Path], Type[Edge]], ColumnElement]
        ] = []
        self.stepped_over_functions: Set[str] = set()
        self.max_length: int = 0
        self.is_reversed = False
        self.start_configuration: Optional[CTE] = None

    def starting_at(self: PB, *start_predicates: Callable[[Type[Node]], ColumnElement]) -> PB:
        """Set a predicate on nodes which starts the traversal.

        This method takes as argument a callable, which is provided with a Node object of the
        underlying graph during build, and must return a valid where-clause-compatible expression.
        """
        self.start_predicates.extend(start_predicates)
        return self

    def stopping_at(self: PB, *stop_predicates: Callable[[Type[Node]], ColumnElement]) -> PB:
        """Set a predicate on nodes which stops the traversal.

        This method takes as argument a callable, which is provided with a Node object of the
        underlying graph during build, and must return a valid where-clause-compatible expression.
        """
        self.stop_predicates.extend(stop_predicates)
        return self

    def continuing_while(
        self: PB,
        *continue_predicates: Callable[[Type[Path], Type[Edge]], ColumnElement],
        edge_detail: bool = True,
    ) -> PB:
        """Set a predicate on configurations and edges which must be satisfied by all edges on a
        valid traversal.

        This method takes as argument a callable, which is provided with a Path object of the
        PathBuilder's current configuration and an edge object of the underlying graph during build,
        and must return a valid where-clause-compatible expression.

        The optional argument edge_detail specifies whether the predicate inspects properties of the
        edge other than its source, target, or UUID
        """
        if edge_detail:
            self.continue_predicates.extend(continue_predicates)
        else:
            self.no_detail_continue_predicates.extend(continue_predicates)
        return self

    def stepping_over(self: PB, *stepped_over_functions: str) -> PB:
        """Specify a set of functions to step over during control-flow traversal."""
        self.stepped_over_functions |= set(stepped_over_functions)
        return self

    def initial_configuration(self: PB, cte: CTE) -> PB:
        """Provide an initial set of configurations (node, stack, and info) at which to start the
        traversal.

        This method takes as argument a SQLAlchemy common table expression, which must have columns
        "uuid", "stack", and "info" containing the UUID of the start node, the starting stack, and
        an information string (which may be null), respectively.
        """
        self.start_configuration = cte
        return self

    @staticmethod
    def build_conjoined_predicate(predicates: List[Callable]) -> Optional[Callable]:
        if len(predicates) == 0:
            return None
        elif len(predicates) == 1:
            return predicates[0]
        else:

            def conjoined_predicate(*pivot: Any) -> Any:
                nonlocal predicates
                conjunction = predicates[0](*pivot)
                for predicate in predicates[1:]:
                    conjunction &= predicate(*pivot)
                return conjunction

            return conjoined_predicate

    def limited_to(self: PB, max_length: int) -> PB:
        """Limit the maximum length of the traversal."""

        mate_assert(max_length >= 0)
        self.max_length = max_length
        return self

    def reverse(self: PB) -> PB:
        """Reverse the direction of traversal.

        This causes the traversal to be jump from target -> source rather than from source -> target; edge IDs in any resulting trace must be interpreted as such.

        Repeated reversals will undo each other.
        """
        self.is_reversed = not self.is_reversed
        return self

    # NOTE: The parameters to this function are also mentioned in the
    # documentation in _make_path_table, so updates here should be reflected
    # there.
    #
    # TODO(#1462): All of these bools should default to False.
    def build(
        self,
        graph: Graph,
        *,
        keep_start: bool = True,
        keep_trace: bool = False,
        keep_edge: bool = False,
        exploration_bound: int = MATE_DEFAULT_EXPLORATION_BOUND,
    ) -> Type[Path]:
        """Concretize the path set defined by this builder, with respect to the given graph.

        :param keep_start: record the starting node for each path
        :param keep_trace: record the sequence of edges visited by the path
        :param keep_edge: record the last edge visited by the path
        :param exploration_bound: the maximum number of path configurations
          to visit during exploration (not a bound on path length)

        All of the ``keep_*`` options have some performance cost if enabled.
        ``keep_trace`` is especially expensive.

        Returns a new Path class, mapped to the view represeting the path set.
        """
        name = uuid.uuid4().hex
        path_table = _make_path_table(f"paths_{name}", graph.Node.__table__, graph.Edge.__table__)

        path_class: Type[Path] = type(
            f"Path_{name}",
            (self.PathBase,),
            {
                "__table__": path_table,
                "stepped_over_functions": self.stepped_over_functions,
                "is_reversed": self.is_reversed,
            },
        )
        mapper(path_class, path_table)

        if self.PathBase.cfl:
            path_class.populate_transition_table(graph)
            path_class.populate_edge_symbol_table(graph)
            assert path_class.transition_table is not None
            assert path_class.edge_symbol_table is not None

        start_predicate = self.build_conjoined_predicate(self.start_predicates)
        stop_predicate = self.build_conjoined_predicate(self.stop_predicates)
        continue_predicate = self.build_conjoined_predicate(self.continue_predicates)
        no_detail_continue_predicate = self.build_conjoined_predicate(
            self.no_detail_continue_predicates
        )

        start_predicate = (lambda _Node: true()) if start_predicate is None else start_predicate
        stop_predicate = (lambda _Node: true()) if stop_predicate is None else stop_predicate
        no_detail_continue_predicate = (
            (lambda _Config, _Edge: true())
            if no_detail_continue_predicate is None
            else no_detail_continue_predicate
        )

        # Project out the appropriate set of columns from the base join, and add
        # on the trace column with the initial edge.
        if self.start_configuration is None:
            base_step = (
                select(
                    [
                        null().label("info"),
                        graph.Node.uuid.label("source")
                        if (keep_start or self.is_reversed)
                        else null().cast(String).label("source"),
                        path_class.initial_stack().label("source_stack")
                        if keep_start and self.PathBase.cfl
                        else null().label("source_stack"),
                        graph.Node.uuid.label("target")
                        if (keep_start or not self.is_reversed)
                        else null().cast(String).label("target"),
                        array([]).cast(ARRAY(String)).label("trace"),
                        path_class.initial_state().label("state"),
                        path_class.initial_stack().label("stack"),
                        Grouping(path_class.initial_stack().cast(ARRAY(String)))[1].label(
                            "stack_top"
                        ),
                        null().cast(String).label("edge"),
                    ]
                )
                .select_from(graph.Node)
                .where(
                    start_predicate(graph.Node)
                    if not self.is_reversed
                    else stop_predicate(graph.Node)
                )
                .cte(recursive=True, name="base_select")
            )
        else:
            base_step = (
                select(
                    [
                        self.start_configuration.c.info.label("info"),
                        graph.Node.uuid.label("source")
                        if (keep_start or self.is_reversed)
                        else null().cast(String).label("source"),
                        self.start_configuration.c.stack.label("source_stack")
                        if keep_start and self.PathBase.cfl
                        else null().label("source_stack"),
                        graph.Node.uuid.label("target")
                        if (keep_start or not self.is_reversed)
                        else null().cast(String).label("target"),
                        array([]).cast(ARRAY(String)).label("trace"),
                        path_class.initial_state().label("state"),
                        self.start_configuration.c.stack.label("stack"),
                        Grouping(self.start_configuration.c.stack)[1].label("stack_top"),
                        null().cast(String).label("edge"),
                    ]
                )
                .select_from(
                    join(
                        graph.Node,
                        self.start_configuration,
                        (graph.Node.uuid == self.start_configuration.c.uuid),
                    )
                )
                .where(
                    start_predicate(graph.Node)
                    if not self.is_reversed
                    else stop_predicate(graph.Node)
                )
                .cte(recursive=True, name="base_select")
            )

        if self.max_length == 0:
            logger.warn("Building paths with unbounded maximum length.")
            max_length_predicate = true()
        else:
            max_length_predicate = func.cardinality(base_step.c.trace) < self.max_length

        if self.PathBase.cfl:
            assert path_class.edge_symbol_table is not None

            if self.is_reversed:
                extension_direction_predicate_1 = (
                    base_step.c.source == path_class.edge_symbol_table.c.target
                )
            else:
                extension_direction_predicate_1 = (
                    base_step.c.target == path_class.edge_symbol_table.c.source
                )

            extension_join_1 = join(
                base_step,
                path_class.edge_symbol_table,
                extension_direction_predicate_1 & max_length_predicate,
            )

            extension_join_1 = join(
                extension_join_1,
                path_class.transition_table,
                (path_class.edge_symbol_table.c.symbol == path_class.transition_table.c.input)  # type: ignore[union-attr]
                & (base_step.c.state == path_class.transition_table.c.old_state)  # type: ignore[union-attr]
                & (base_step.c.stack_top == path_class.transition_table.c.old_stack)  # type: ignore[union-attr]
                & no_detail_continue_predicate(base_step, path_class.edge_symbol_table.c),
            )

            if continue_predicate is not None:
                extension_join_1 = join(
                    extension_join_1,
                    graph.Edge.__table__,
                    (path_class.edge_symbol_table.c.uuid == graph.Edge.uuid)
                    & continue_predicate(base_step, graph.Edge),
                )
        else:
            if self.is_reversed:
                extension_direction_predicate_1 = base_step.c.source == graph.Edge.target
            else:
                extension_direction_predicate_1 = base_step.c.target == graph.Edge.source

            if continue_predicate is not None:
                extension_join_1 = join(
                    base_step,
                    graph.Edge.__table__,
                    extension_direction_predicate_1
                    & no_detail_continue_predicate(base_step, graph.Edge)
                    & continue_predicate(base_step, graph.Edge)
                    & max_length_predicate,
                )
            else:
                extension_join_1 = join(
                    base_step,
                    graph.Edge.__table__,
                    extension_direction_predicate_1
                    & no_detail_continue_predicate(base_step, graph.Edge)
                    & max_length_predicate,
                )

        if self.PathBase.cfl:
            edge_rel = path_class.edge_symbol_table
        else:
            edge_rel = graph.Edge.__table__

        assert edge_rel is not None

        if not self.is_reversed:
            source_columns = [base_step.c.source]
            target_columns = [
                edge_rel.c.target.label("target"),
            ]
            augmented_trace = func.array_append(base_step.c.trace, edge_rel.c.uuid).label("trace")
        else:
            source_columns = [
                edge_rel.c.source.label("source"),
            ]
            target_columns = [base_step.c.target]
            augmented_trace = func.array_prepend(edge_rel.c.uuid, base_step.c.trace).label("trace")

        # Project out the appropriate set of columns from the first extension
        # join, and augment each path trace with the UUID of the edge just
        # added.
        extension_select_1 = select(
            [
                base_step.c.info,
                *source_columns,
                base_step.c.source_stack,
                *target_columns,
                augmented_trace if keep_trace else base_step.c.trace,
                path_class.transition_table.c.new_state.label("state")  # type: ignore[union-attr]
                if self.PathBase.cfl
                else base_step.c.state,
                case(
                    [
                        (
                            (
                                path_class.transition_table.c.new_stack[  # type: ignore[union-attr]
                                    func.cardinality(path_class.transition_table.c.new_stack)  # type: ignore[union-attr]
                                ]
                                == literal(BOT)
                            ),
                            path_class.transition_table.c.new_stack,  # type: ignore[union-attr]
                        )
                    ],
                    else_=func.array_cat(
                        path_class.transition_table.c.new_stack,  # type: ignore[union-attr]
                        base_step.c.stack[2 : func.cardinality(base_step.c.stack)],
                    ),
                )
                .cast(ARRAY(String))
                .label("stack")
                if self.PathBase.cfl
                else base_step.c.stack,
                case(
                    [
                        (
                            func.cardinality(path_class.transition_table.c.new_stack) == 0,  # type: ignore[union-attr]
                            base_step.c.stack[2],
                        ),
                    ],
                    else_=Grouping(path_class.transition_table.c.new_stack)[1],  # type: ignore[union-attr]
                )
                .cast(String)
                .label("stack_top")
                if self.PathBase.cfl
                else base_step.c.stack_top,
                edge_rel.c.uuid.label("edge") if keep_edge else base_step.c.edge,
            ],
        ).select_from(extension_join_1)

        # TODO: handle continuing while predicates with Node attributes
        path_steps = (
            base_step.union(extension_select_1).alias()
            if exploration_bound is None
            else base_step.union(extension_select_1).select().limit(exploration_bound).alias()
        )

        if (self.stop_predicates != [] and not self.is_reversed) or (
            self.start_predicates != [] and self.is_reversed
        ):
            filtered_paths = join(
                path_steps,
                graph.Node,
                (
                    (path_steps.c.target == graph.Node.uuid)
                    if not self.is_reversed
                    else (path_steps.c.source == graph.Node.uuid)
                )
                & (
                    stop_predicate(graph.Node)
                    if not self.is_reversed
                    else start_predicate(graph.Node)
                ),
            )
        else:
            filtered_paths = path_steps

        # Issue the VIEW definition. This statement should be instantaneous.
        graph.session.bind.execute(CreateView(f"paths_{name}", filtered_paths.select()))

        return path_class


def _make_edge_symbol_table(name: str, edge_table: Table) -> Table:
    uuid_comment = "UUID of the edge in the CPG"
    source_comment = "UUID of the edge's source node"
    target_comment = "UUID of the edge's target node"
    symbol_comment = "A symbol encoding the relevant information about the edge that will be used in the PDA's transition relation"
    return Table(
        f"{name}_edge_symbols",
        _metadata,
        Column(
            "uuid", String, primary_key=True, nullable=False, doc=uuid_comment, comment=uuid_comment
        ),
        Column(
            "source", String, index=True, nullable=False, doc=source_comment, comment=source_comment
        ),
        Column(
            "target", String, index=True, nullable=False, doc=target_comment, comment=target_comment
        ),
        Column(
            "symbol",
            String,
            primary_key=True,
            index=True,
            nullable=False,
            doc=symbol_comment,
            comment=symbol_comment,
        ),
        ForeignKeyConstraint(
            ["uuid"],
            [edge_table.c.uuid],
        ),
        Index(f"{name}_sym_ss", "symbol", "source"),
        Index(f"{name}_sym_st", "symbol", "target"),
        extend_existing=True,
    )


def _make_transition_table(name: str) -> Table:
    old_state_comment = "The PDA state for which this transition is applicable"
    old_stack_comment = "The top item of the PDA stack for which this transition is applicable (always a single symbol, even though the type is array)"
    input_comment = "The edge symbol that must match for this transition to fire"
    new_state_comment = "The state to transition to"
    new_stack_comment = "A stack of values to replace the top item on the stack with, or a replacement for the entire stack if new_stack ends with '$'"
    return Table(
        f"{name}_rules",
        _metadata,
        Column(
            "old_state", String, nullable=False, doc=old_state_comment, comment=old_state_comment
        ),
        Column(
            "old_stack",
            String,
            nullable=False,
            doc=old_stack_comment,
            comment=old_stack_comment,
        ),
        Column("input", String, nullable=False, doc=input_comment, comment=input_comment),
        Column(
            "new_state", String, nullable=False, doc=new_state_comment, comment=new_state_comment
        ),
        Column(
            "new_stack",
            ARRAY(String),
            nullable=False,
            doc=new_stack_comment,
            comment=new_stack_comment,
        ),
        Index(f"{name}_rules_j", "old_state", "old_stack", "input"),
        extend_existing=True,
    )



"""This module contains functions which return lists of names of known library functions that may be
interesting for security or dataflow."""

from typing import List

_sql_keywords: List[str] = ["ALTER", "SELECT", "TABLE", "CREATE", "DELETE", "DROP", "INSERT"]

_http_keywords: List[str] = ["url", "POST", "GET", "REQUEST", "http", "https"]

_os_commands: List[str] = ["find ", "exec ", "source ", "ssh "]

_string_output_func_names: List[str] = [
    "fprintf",
    "printf",
    "snprintf",
    "sprintf",
    "dprintf",
    "vfprintf",
    "vprintf",
    "vdprintf",
    "vsprintf",
    "vsnprintf",
    "vsscanf",
    "asprintf",
    "wprintf",
    "fwprintf",
    "swprintf",
    "vwprintf",
    "vfwprintf",
    "vswprintf",
    "snprintf",
    "fputs",
    "puts",
    "write",
    "fwrite",
    "fputws",
]


def formatted_output_function_names() -> List[str]:
    """Returns the names of functions like printf which format and print strings."""
    return _string_output_func_names


def protocol_keywords() -> List[str]:
    """Returns keywords of known protocols like SQL and HTTP."""
    return _sql_keywords + _http_keywords


def sql_keywords() -> List[str]:
    """Returns SQL keywords."""
    return _sql_keywords + [keyword.lower() for keyword in _sql_keywords]



"""This module contains a class ``NodeMixin`` which is equal to ``mate_query.cpg.models.core.Node``
at type-checking time, but is just ``object`` at runtime.

This is used as a base class for all the node mix-ins, because they are only mixed into subclasses
of ``Node``. Mypy can't use that information without a little bit of help.
"""

from typing import TYPE_CHECKING

from mate_common.utils import SPHINX_BUILD

if TYPE_CHECKING or SPHINX_BUILD:
    from mate_query.cpg.models.core.node import Node

    class NodeMixin(Node):
        pass

else:

    class NodeMixin:
        pass



"""``mate_query``-specific configuration state and settings."""

from typing import Final

from mate_common.config import get_config

# Default limit for the number of entries to explore during path queries
# made by mate_server / flowfinder endpoints.
MATE_SERVER_EXPLORATION_BOUND: Final[int] = get_config(
    "MATE_SERVER_EXPLORATION_BOUND", int, 5_000_000
)

# Default limit for the number of entries to explore during path queries.
# Chosen empirically by seeing what works well on challenge problems...
# TODO(sm): tune this further
MATE_DEFAULT_EXPLORATION_BOUND: Final[int] = get_config(
    "MATE_DEFAULT_EXPLORATION_BOUND", int, 5_000_000
)

MATE_STORAGE_ACCESS_KEY: Final[str] = get_config("MATE_STORAGE_ACCESS_KEY", str, "MATE_ACCESS_KEY")

MATE_STORAGE_SECRET_KEY: Final[str] = get_config("MATE_STORAGE_SECRET_KEY", str, "MATE_SECRET_KEY")
